from __future__ import annotations

from collections.abc import Callable
import os
import pathlib
from typing import Literal

from prettyqt import constants, core, gui
from prettyqt.qt import QtGui
from prettyqt.utils import InvalidParamError, bidict, datatypes, get_repr


ANCESTER_MODES = bidict(
    exclude_transients=QtGui.QWindow.AncestorMode.ExcludeTransients,
    include_transients=QtGui.QWindow.AncestorMode.IncludeTransients,
)

VISIBILITY = bidict(
    windowed=QtGui.QWindow.Visibility.Windowed,
    minimized=QtGui.QWindow.Visibility.Minimized,
    maximized=QtGui.QWindow.Visibility.Maximized,
    fullscreen=QtGui.QWindow.Visibility.FullScreen,
    automatic=QtGui.QWindow.Visibility.AutomaticVisibility,
    hidden=QtGui.QWindow.Visibility.Hidden,
)

VisibilityStr = Literal[
    "windowed", "minimized", "maximized", "fullscreen", "automatic", "hidden"
]


class WindowMixin(core.ObjectMixin, gui.SurfaceMixin):
    def __repr__(self):
        return get_repr(self)

    # def serialize_fields(self):
    #     return dict(
    #         speed=self.speed(),
    #         visibility=self.get_visibility(),
    #         scaled_size=self.scaledSize(),
    #         background_color=self.backgroundColor(),
    #     )

    def add_shortcut(
        self,
        keysequence: datatypes.KeyCombinationType,
        callback: Callable | None = None,
        context: constants.ShortcutContextStr = "window",
    ) -> gui.Shortcut:
        if not isinstance(keysequence, QtGui.QKeySequence):
            keysequence = gui.KeySequence(keysequence)
        context = constants.SHORTCUT_CONTEXT[context]
        return gui.Shortcut(keysequence, self, callback, context=context)

    def set_visibility(self, visibility: VisibilityStr):
        """Set window visibility.

        Args:
            visibility: window visibility

        Raises:
            InvalidParamError: window visibility does not exist
        """
        if visibility not in VISIBILITY:
            raise InvalidParamError(visibility, VISIBILITY)
        self.setVisibility(VISIBILITY[visibility])

    def get_visibility(self) -> VisibilityStr:
        """Get the current window visibility.

        Returns:
            window visibility
        """
        return VISIBILITY.inverse[self.visibility()]

    def start_system_resize(self, edge: constants.EdgeStr) -> bool:
        """Start system resize.

        Args:
            edge: edge to resize

        Raises:
            InvalidParamError: edge does not exist
        """
        if edge not in constants.EDGES:
            raise InvalidParamError(edge, constants.EDGES)
        return self.startSystemResize(constants.EDGES[edge])

    def get_screen(self) -> gui.Screen:
        return gui.Screen(self.screen())

    def get_cursor(self) -> gui.Cursor:
        return gui.Cursor(self.cursor())

    def get_icon(self) -> gui.Icon | None:
        icon = self.icon()
        return None if icon.isNull() else gui.Icon(icon)

    def set_modality(self, modality: constants.WindowModalityStr) -> None:
        """Set modality for the window.

        Args:
            modality: modality for the window

        Raises:
            InvalidParamError: modality type does not exist
        """
        if modality not in constants.WINDOW_MODALITY:
            raise InvalidParamError(modality, constants.WINDOW_MODALITY)
        self.setModality(constants.WINDOW_MODALITY[modality])

    def get_modality(self) -> constants.WindowModalityStr:
        return constants.WINDOW_MODALITY.inverse[self.modality()]

    def set_file_path(self, file_path: os.PathLike):
        path = os.fspath(file_path)
        self.setFilePath(path)

    def get_file_path(self) -> pathlib.Path:
        return pathlib.Path(self.filePath())

    def get_type(self) -> constants.WindowTypeStr:
        return constants.WINDOW_TYPE.inverse[self.type()]

    def get_window_state(self) -> constants.WindowStateStr:
        return constants.WINDOW_STATES.inverse[self.windowState()]

    def get_window_states(self) -> list[constants.WindowStateStr]:
        return constants.WINDOW_STATES.get_list(self.windowStates())


class Window(WindowMixin, QtGui.QWindow):
    pass


if __name__ == "__main__":
    wnd = Window()



from __future__ import annotations

from typing import Literal

from prettyqt import bluetooth, core
from prettyqt.qt import QtBluetooth
from prettyqt.utils import bidict


DISCOVERY_METHODS = bidict(
    none=QtBluetooth.QBluetoothDeviceDiscoveryAgent.DiscoveryMethod.NoMethod,
    classic=QtBluetooth.QBluetoothDeviceDiscoveryAgent.DiscoveryMethod.ClassicMethod,
    low_energy=QtBluetooth.QBluetoothDeviceDiscoveryAgent.DiscoveryMethod.LowEnergyMethod,
)

DiscoveryMethodStr = Literal["none", "classic", "low_energy"]

Error = QtBluetooth.QBluetoothDeviceDiscoveryAgent.Error

ERROR = bidict(
    none=Error.NoError,
    powered_off=Error.PoweredOffError,
    input_output=Error.InputOutputError,
    invalid_bluetooth_adapter=Error.InvalidBluetoothAdapterError,
    unsupported_platform=Error.UnsupportedPlatformError,
    unsupported_discovery=Error.UnsupportedDiscoveryMethod,
    unknown=Error.UnknownError,
)

ErrorStr = Literal[
    "none",
    "powered_off",
    "input_output",
    "invalid_bluetooth_adapter",
    "unsupported_platform",
    "unsupported_discovery",
    "unknown",
]

# INQUIRY_TYPES = bidict(
#     unlimited=QtBluetooth.QBluetoothDeviceDiscoveryAgent.GeneralUnlimitedInquiry,
#     limited=QtBluetooth.QBluetoothDeviceDiscoveryAgent.LimitedInquiry,
# )

# InquiryTypeStr = Literal["unlimited", "limited"]


class BluetoothDeviceDiscoveryAgent(
    core.ObjectMixin, QtBluetooth.QBluetoothDeviceDiscoveryAgent
):
    device_discovered = core.Signal(object)  # bluetooth.BluetoothDeviceInfo casts signal.

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.deviceDiscovered.connect(self._on_device_discovered)

    def _on_device_discovered(self, info):
        info = bluetooth.BluetoothDeviceInfo(info)
        self.device_discovered.emit(info)

    # def set_inquiry_type(self, typ: InquiryTypeStr):
    #     """Set inquiry type.

    #     Args:
    #         typ: inquiry type

    #     Raises:
    #         InvalidParamError: inquiry type does not exist
    #     """
    #     if typ not in INQUIRY_TYPES:
    #         raise InvalidParamError(typ, INQUIRY_TYPES)
    #     self.setInquiryType(INQUIRY_TYPES[typ])

    # def get_inquiry_type(self) -> InquiryTypeStr:
    #     """Get the current inquiry type.

    #     Returns:
    #         inquiry type
    #     """
    #     return INQUIRY_TYPES.inverse[self.inquiryType()]

    def start_discovery(self, classic: bool = True, low_energy: bool = True):
        flag = QtBluetooth.QBluetoothDeviceDiscoveryAgent.NoMethod
        if classic:
            flag |= QtBluetooth.QBluetoothDeviceDiscoveryAgent.ClassicMethod
        if low_energy:
            flag |= QtBluetooth.QBluetoothDeviceDiscoveryAgent.LowEnergyMethod
        self.start(flag)

    def get_error(self) -> ErrorStr:
        return ERROR.inverse[self.error()]

    def get_supported_discovery_methods(self) -> list[DiscoveryMethodStr]:
        return DISCOVERY_METHODS.get_list(self.supportedDiscoveryMethods())


if __name__ == "__main__":
    import logging

    from prettyqt import core

    app = core.app()
    agent = BluetoothDeviceDiscoveryAgent(app)
    agent.setLowEnergyDiscoveryTimeout(500)
    agent.device_discovered.connect(lambda x: logging.info(x))
    agent.finished.connect(app.quit)
    agent.start_discovery()
    with app.debug_mode():
        app.main_loop()



from __future__ import annotations

from typing import Literal

from prettyqt import constants, gui
from prettyqt.qt import QtGui
from prettyqt.utils import InvalidParamError, bidict


LINE_HEIGHT_TYPES = bidict(
    single=QtGui.QTextBlockFormat.LineHeightTypes.SingleHeight,
    proportional=QtGui.QTextBlockFormat.LineHeightTypes.ProportionalHeight,
    fixed=QtGui.QTextBlockFormat.LineHeightTypes.FixedHeight,
    minimum=QtGui.QTextBlockFormat.LineHeightTypes.MinimumHeight,
    line_distance=QtGui.QTextBlockFormat.LineHeightTypes.LineDistanceHeight,
)

LineHeightTypeStr = Literal[
    "single",
    "proportional",
    "fixed",
    "minimum",
    "line_distance",
]

MARKER_TYPE = bidict(
    none=QtGui.QTextBlockFormat.MarkerType.NoMarker,
    unchecked=QtGui.QTextBlockFormat.MarkerType.Unchecked,
    checked=QtGui.QTextBlockFormat.MarkerType.Checked,
)

MarkerTypeStr = Literal[
    "none",
    "unchecked",
    "checked",
]


class TextBlockFormat(gui.TextFormatMixin, QtGui.QTextBlockFormat):
    def set_marker(self, marker: MarkerTypeStr):
        """Set the marker.

        Args:
            marker: marker

        Raises:
            InvalidParamError: invalid marker
        """
        if marker not in MARKER_TYPE:
            raise InvalidParamError(marker, MARKER_TYPE)
        self.setMarker(MARKER_TYPE[marker])

    def get_marker(self) -> MarkerTypeStr:
        """Get current marker.

        Returns:
            current marker
        """
        return MARKER_TYPE.inverse[self.marker()]

    def set_alignment(self, alignment: constants.AlignmentStr):
        """Set the alignment of the format.

        Args:
            alignment: alignment for the format

        Raises:
            InvalidParamError: alignment does not exist
        """
        if alignment not in constants.ALIGNMENTS:
            raise InvalidParamError(alignment, constants.ALIGNMENTS)
        self.setAlignment(constants.ALIGNMENTS[alignment])

    def get_alignment(self) -> constants.AlignmentStr:
        """Return current alignment.

        Returns:
            alignment
        """
        return constants.ALIGNMENTS.inverse[self.alignment()]

    def set_page_break_policy(self, policy: gui.textformat.PageBreakFlagStr):
        """Set page break policy.

        Args:
            policy: page break policy

        Raises:
            InvalidParamError: page break policy does not exist
        """
        if policy not in gui.textformat.PAGE_BREAK_FLAG:
            raise InvalidParamError(policy, gui.textformat.PAGE_BREAK_FLAG)
        self.setPageBreakPolicy(gui.textformat.PAGE_BREAK_FLAG[policy])

    def get_page_break_policy(self) -> gui.textformat.PageBreakFlagStr:
        """Get the current page break policy.

        Returns:
            page break policy
        """
        return gui.textformat.PAGE_BREAK_FLAG.inverse[self.pageBreakPolicy()]



from __future__ import annotations

from collections.abc import Iterator

from typing_extensions import Self

from prettyqt import constants, widgets
from prettyqt.qt import QtCore, QtWidgets
from prettyqt.utils import InvalidParamError, datatypes


class SplitterMixin(widgets.FrameMixin):
    def __init__(
        self,
        orientation: (constants.OrientationStr | QtCore.Qt.Orientation) = "horizontal",
        **kwargs,
    ):
        if isinstance(orientation, QtCore.Qt.Orientation):
            ori = orientation
        else:
            ori = constants.ORIENTATION[orientation]
        self._next_container = None
        self._stack = []
        super().__init__(ori, **kwargs)
        self.setHandleWidth(10)

    def __enter__(self):
        if self._next_container is not None:
            self._next_container.__enter__()
            self._stack.append(self._next_container)
            self._next_container = None
        return self

    def __exit__(self, *_):
        if self._stack:
            item = self._stack.pop()
            item.__exit__()

    def __getitem__(self, index: int | str) -> QtWidgets.QWidget:
        if isinstance(index, int):
            return self.widget(index)
        result = self.find_child(QtWidgets.QWidget, index)
        if result is None:
            raise KeyError("Widget not found")
        return result

    def __setitem__(self, index: int, value: QtWidgets.QWidget):
        self.replaceWidget(index, value)

    def __iter__(self) -> Iterator[QtWidgets.QWidget]:
        return iter(self.get_children())

    def __len__(self) -> int:
        return self.count()

    def __contains__(self, item: QtWidgets.QWidget):
        return self.indexOf(item) >= 0

    def __add__(self, other: QtWidgets.QWidget | QtWidgets.QLayout):
        self.add(other)
        return self

    def __iadd__(self, item):
        self.add(item)
        return self

    def createHandle(self) -> widgets.SplitterHandle:
        return widgets.SplitterHandle(self.orientation(), self)

    # def saveState(self):
    #     sizes = self.sizes()
    #     if all(x == 0 for x in sizes):
    #         sizes = [10] * len(sizes)
    #     return {'sizes': sizes}

    # def restoreState(self, state):
    #     sizes = state['sizes']
    #     self.setSizes(sizes)
    #     for i in range(len(sizes)):
    #         self.setStretchFactor(i, sizes[i])

    @classmethod
    def create(
        cls,
        parent: QtWidgets.QWidget | QtWidgets.QLayout,
        margins: datatypes.MarginsType | None = None,
        orientation: constants.OrientationStr | None = None,
        **kwargs,
    ):
        if orientation in constants.ORIENTATION:
            orientation = constants.ORIENTATION[orientation]
        match parent:
            case QtWidgets.QWidget():
                new = cls(parent=parent, orientation=orientation, **kwargs)
                widgets.HBoxLayout.create(parent).add(new)  # .create ?
            case QtWidgets.QLayout():
                new = cls(orientation, **kwargs)
                parent.add(new)
        if margins:
            new.setContentsMargins(*margins)

        if orientation:
            new.setOrientation(orientation)
        new._stack = []
        new._next_container = None
        return new

    def get_sub_layout(self, layout: str, *args, **kwargs):
        match layout:
            case "horizontal":
                self._next_container = widgets.HBoxLayout.create(
                    self._layout, *args, **kwargs
                )
            case "vertical":
                self._next_container = widgets.VBoxLayout.create(
                    self._layout, *args, **kwargs
                )
            case "grid":
                self._next_container = widgets.GridLayout.create(
                    self._layout, *args, **kwargs
                )
            case "form":
                self._next_container = widgets.FormLayout.create(
                    self._layout, *args, **kwargs
                )
            case "stacked":
                self._next_container = widgets.StackedLayout.create(
                    self._layout, *args, **kwargs
                )
            case "flow":
                from prettyqt import custom_widgets

                self._next_container = custom_widgets.FlowLayout.create(
                    self._layout, *args, **kwargs
                )
            case "splitter":
                self._next_container = widgets.Splitter.create(
                    self._layout, *args, **kwargs
                )
            case "scroll":
                self._next_container = widgets.ScrollArea.create(
                    self._layout, *args, **kwargs
                )
            case _:
                raise ValueError("Invalid Layout")
        return self

    @property
    def _layout(self):
        return self._stack[-1] if self._stack else self

    def get_children(self) -> list[QtWidgets.QWidget]:
        return [self[i] for i in range(self.count())]

    def add_widget(
        self,
        widget: QtWidgets.QWidget,
        stretch: int | None = None,
        collapsible: bool = True,
        position: int | None = None,
    ) -> QtWidgets.QWidget:
        if position is None:
            self.addWidget(widget)
        else:
            self.insertWidget(position, widget)
        index = self.indexOf(widget)
        if stretch is not None:
            self.setStretchFactor(index, stretch)
        self.setCollapsible(index, collapsible)
        return widget

    def add_layout(
        self,
        layout: QtWidgets.QLayout,
        stretch: int | None = None,
        collapsible: bool = True,
        position: int | None = None,
    ) -> widgets.Widget:
        widget = widgets.Widget()
        widget.set_layout(layout)
        return self.add_widget(
            widget, stretch=stretch, collapsible=collapsible, position=position
        )

    def add(
        self,
        item: QtWidgets.QWidget
        | QtWidgets.QLayout
        | list[QtWidgets.QWidget | QtWidgets.QLayout],
        stretch: int | None = None,
    ):
        match item:
            case QtWidgets.QWidget():
                self.add_widget(item, stretch=stretch)
            case QtWidgets.QLayout():
                self.add_layout(item)
            case list():
                for i in item:
                    self.add(i)
        return item

    @classmethod
    def from_widgets(
        cls,
        *widgets: QtWidgets.QWidget,
        horizontal: bool = False,
        parent: QtWidgets.QWidget | None = None,
    ) -> Self:
        splitter = cls("horizontal" if horizontal else "vertical", parent=parent)
        for widget in widgets:
            splitter += widget
        return splitter

    def set_orientation(self, orientation: constants.OrientationStr):
        """Set the orientation of the splitter.

        Args:
            orientation: orientation for the splitter

        Raises:
            InvalidParamError: orientation does not exist
        """
        if orientation not in constants.ORIENTATION:
            raise InvalidParamError(orientation, constants.ORIENTATION)
        self.setOrientation(constants.ORIENTATION[orientation])

    def get_orientation(self) -> constants.OrientationStr:
        """Return current orientation.

        Returns:
            orientation
        """
        return constants.ORIENTATION.inverse[self.orientation()]


class Splitter(SplitterMixin, QtWidgets.QSplitter):
    pass


if __name__ == "__main__":
    app = widgets.app()
    widget = Splitter()
    widget.show()
    app.main_loop()



from __future__ import annotations

from collections.abc import Iterator
import html
import os
from typing import Any

from prettyqt import constants, core, gui, iconprovider
from prettyqt.qt import QtCore, QtGui
from prettyqt.utils import InvalidParamError, datatypes, get_repr, serializemixin


class StandardItem(serializemixin.SerializeMixin, QtGui.QStandardItem):
    def __repr__(self):
        return get_repr(self, self.get_icon(), self.text())

    def serialize_fields(self):
        return dict(
            text=self.text(),
            tool_tip=self.toolTip(),
            status_tip=self.statusTip(),
            icon=self.get_icon(),
            data=self.data(),
        )

    def __getitem__(
        self, index: int | tuple[int, int] | QtCore.QModelIndex
    ) -> QtGui.QStandardItem:
        match index:
            case int():
                return self.child(index)
            case tuple():
                return self.child(*index)
            case _:
                raise KeyError(index)

    def __delitem__(self, index: int | tuple[int, int]):
        item = self.takeRow(index) if isinstance(index, int) else self.takeChild(*index)
        if item is None:
            raise KeyError(index)
        return item

    def __iter__(self) -> Iterator[QtGui.QStandardItem]:
        return iter(self.get_children())

    def __add__(self, other: str | QtGui.QStandardItem) -> StandardItem:
        if isinstance(other, QtGui.QStandardItem | str):
            self.add(other)
            return self
        raise TypeError("wrong type for addition")

    def get_children(self) -> list[QtGui.QStandardItem]:
        return [self.child(index) for index in range(self.rowCount())]

    def add(self, *item: str | QtGui.QStandardItem):
        for i in item:
            new_item = gui.StandardItem(i) if isinstance(i, str) else i
            self.appendRow([new_item])

    def clone(self):
        item = type(self)()
        core.DataStream.copy_data(self, item)
        assert type(item) == StandardItem
        return item

    def set_icon(self, icon: datatypes.IconType):
        """Set the icon for the action.

        Args:
            icon: icon to use
        """
        icon = iconprovider.get_icon(icon)
        self.setIcon(icon)

    def set_checkstate(self, state: constants.CheckStateStr):
        """Set checkstate of the checkbox.

        Args:
            state: checkstate to use

        Raises:
            InvalidParamError: invalid checkstate
        """
        if state not in constants.CHECK_STATE:
            raise InvalidParamError(state, constants.CHECK_STATE)
        self.setCheckState(constants.CHECK_STATE[state])

    def get_checkstate(self) -> constants.CheckStateStr:
        """Return checkstate.

        Returns:
            checkstate
        """
        return constants.CHECK_STATE.inverse[self.checkState()]

    def get_background(self) -> gui.Brush:
        return gui.Brush(self.background())

    def get_foreground(self) -> gui.Brush:
        return gui.Brush(self.foreground())

    def get_font(self) -> gui.Font:
        return gui.Font(self.font())

    def get_icon(self) -> gui.Icon | None:
        icon = self.icon()
        return None if icon.isNull() else gui.Icon(icon)

    def set_data(self, data: Any, role: constants.ItemDataRoleStr | int):
        if isinstance(role, str):
            role = constants.ITEM_DATA_ROLE[role]
        super().setData(data, role)

    def get_data(self, role: constants.ItemDataRoleStr | int):
        if isinstance(role, str):
            role = constants.ITEM_DATA_ROLE[role]
        return super().data(role)

    def set_tooltip(
        self,
        tooltip: str | datatypes.PathType,
        size: datatypes.SizeType | None = None,
        rich_text: bool = False,
    ):
        if isinstance(tooltip, os.PathLike):
            path = os.fspath(tooltip)
            if size is None:
                tooltip = f"<img src={path!r}>"
            else:
                if isinstance(size, QtCore.QSize):
                    size = (size.width(), size.height())
                tooltip = f'<img src={path!r} width="{size[0]}" height="{size[1]}">'
        if rich_text:
            tooltip = f"<html>{html.escape(tooltip)}</html>"
        super().setToolTip(tooltip)

    def set_size_hint(self, hint: datatypes.SizeType):
        if isinstance(hint, tuple):
            hint = QtCore.QSize(*hint)
        self.setSizeHint(hint)

    def add_item(
        self,
        name: str = "",
        icon: datatypes.IconType = None,
        data: dict | None = None,
        foreground: QtGui.QBrush | None = None,
        background: QtGui.QBrush | None = None,
        font: QtGui.QFont | None = None,
        selectable: bool = True,
        enabled: bool = True,
        editable: bool = False,
        status_tip: str | None = None,
        tool_tip: str | None = None,
        whats_this: str | None = None,
        # text_alignment: Optional[str] = None,
        checkstate: constants.CheckStateStr | None = None,
        flags: QtCore.Qt.ItemFlag | None = None,
        size_hint: datatypes.SizeType | None = None,
        is_user_type: bool = False,
    ) -> StandardItem:
        item = StandardItem(name)
        if icon is not None:
            icon = iconprovider.get_icon(icon)
            item.setIcon(icon)
        if data is not None:
            for k, v in data.items():
                item.set_data(v, k)
        if foreground is not None:
            item.setForeground(foreground)
        if background is not None:
            item.setBackground(background)
        if font is not None:
            item.setFont(font)
        if flags is not None:
            item.setFlags(flags)
        if enabled:
            item.setEnabled(enabled)
        if editable:
            item.setEditable(editable)
        if selectable:
            item.setSelectable(selectable)
        if status_tip:
            item.setStatusTip(status_tip)
        if tool_tip:
            item.setToolTip(tool_tip)
        if whats_this:
            item.setWhatsThis(whats_this)
        if size_hint is not None:
            item.set_size_hint(size_hint)
        if checkstate is not None:
            item.set_checkstate(checkstate)
        self.appendRow([item])
        return item


if __name__ == "__main__":
    item = StandardItem()
    item.set_data("test", 1000)
    item2 = StandardItem()
    item.add(item2)
    print(item.child(0))



"""eventfilters module.

Contains custom EventFilter classes
"""

from .baseeventfilter import BaseEventFilter
from .hovericoneventfilter import HoverIconEventFilter
from .animatedtooltipeventfilter import AnimatedToolTipEventFilter
from .eventcatcher import EventCatcher


__all__ = [
    "BaseEventFilter",
    "HoverIconEventFilter",
    "AnimatedToolTipEventFilter",
    "EventCatcher",
]



from __future__ import annotations

from collections import defaultdict
from collections.abc import Sequence
import inspect
import logging

from prettyqt import constants, core
from prettyqt.qt import QtCore, QtWidgets


logger = logging.getLogger(__name__)


def find_common_ancestor(cls_list):
    mros = [list(inspect.getmro(cls)) for cls in cls_list]
    track = defaultdict(int)
    while mros:
        for mro in mros:
            cur = mro.pop(0)
            track[cur] += 1
            if (
                track[cur] >= len(cls_list)
                and not cur.__name__.endswith("Mixin")
                and cur.__name__.startswith("Q")
            ):
                return cur
            if len(mro) == 0:
                mros.remove(mro)
    raise TypeError("Couldnt find common base class")


class WidgetsDetailsModel(core.AbstractTableModel):
    def __init__(self, items: Sequence[QtWidgets.QWidget], **kwargs):
        super().__init__(**kwargs)
        self.items = items
        common_ancestor = find_common_ancestor([type(i) for i in self.items])
        logger.debug(f"{type(self).__name__}: found common ancester {common_ancestor}")
        self.props = core.MetaObject(common_ancestor.staticMetaObject).get_properties()
        self.props.sort(key=lambda x: x.get_name())

    def columnCount(self, parent=None):
        return len(self.props)

    def headerData(
        self,
        section: int,
        orientation: QtCore.Qt.Orientation,
        role: QtCore.Qt.ItemDataRole,
    ) -> str | None:
        match orientation, role, section:
            case constants.VERTICAL, constants.DISPLAY_ROLE, _:
                widget = self.items[section]
                return repr(widget)
            case constants.HORIZONTAL, constants.DISPLAY_ROLE, _:
                return self.props[section].get_name()

    def data(self, index, role=constants.DISPLAY_ROLE):
        if not index.isValid():
            return None
        prop = self.props[index.column()]
        widget = self.items[index.row()]
        match role:
            case constants.DISPLAY_ROLE | constants.EDIT_ROLE:
                return prop.read(widget)
            case constants.USER_ROLE:
                return prop.read(widget)

    def setData(self, index, value, role=constants.DISPLAY_ROLE):
        prop = self.props[index.column()]
        widget = self.items[index.row()]
        match role:
            case constants.USER_ROLE:
                with self.reset_model():
                    prop.write(widget, value)
                # self.update_row(index.row())
                return True
        return False

    def rowCount(self, parent=None):
        """Override for AbstractitemModel base method."""
        parent = parent or core.ModelIndex()
        if parent.column() > 0:
            return 0
        return 0 if parent.isValid() else len(self.items)

    def flags(self, index):
        prop = self.props[index.column()]
        if prop.isWritable():
            return (
                super().flags(index)
                | constants.IS_EDITABLE
                | constants.IS_ENABLED
                | constants.IS_SELECTABLE
            )
        return super().flags(index)


if __name__ == "__main__":
    from prettyqt import widgets
    from prettyqt.custom_delegates import variantdelegate

    app = widgets.app()
    view = widgets.TableView()
    view.set_icon("mdi.folder")
    items = [widgets.TableWidget(), widgets.TableWidget()]
    model = WidgetsDetailsModel(items, parent=view).transpose()
    delegate = variantdelegate.VariantDelegate(parent=view)
    view.set_model(model)
    view.setEditTriggers(view.EditTrigger.AllEditTriggers)
    view.set_delegate(delegate)
    view.resize(1000, 1000)
    with app.debug_mode():
        view.show()
        app.main_loop()



from __future__ import annotations

from prettyqt import constants, core
from prettyqt.qt import QtCore


# TYPES = {
#     bool: 1,
#     int: 2,
#     str: 10,
#     float: 38,
#     QtGui.QColor: 67,
#     QtGui.QCursor: 74,
#     QtCore.QDate: 14,
#     QtCore.QSize: 21,
#     QtCore.QTime: 15,
#     list: 9,
#     QtGui.QPolygon: 71,
#     QtGui.QPolygonF: 86,
#     QtGui.QColor: 67,
#     QtGui.QColorSpace: 87,
#     QtCore.QSizeF: 22,
#     QtCore.QRectF: 20,
#     QtCore.QLine: 23,
#     QtGui.QTextLength: 77,
#     dict: 8,
#     QtGui.QIcon: 69,
#     QtGui.QPen: 76,
#     QtCore.QLineF: 24,
#     QtGui.QTextFormat: 78,
#     QtCore.QRect: 19,
#     QtCore.QPoint: 25,
#     QtCore.QUrl: 17,
#     QtCore.QRegularExpression: 44,
#     QtCore.QDateTime: 16,
#     QtCore.QPointF: 26,
#     QtGui.QPalette: 68,
#     QtGui.QFont: 64,
#     QtGui.QBrush: 66,
#     QtGui.QRegion: 72,
#     QtGui.QImage: 70,
#     QtGui.QKeySequence: 75,
#     QtWidgets.QSizePolicy: 121,
#     QtGui.QPixmap: 65,
#     QtCore.QLocale: 18,
#     QtGui.QBitmap: 73,
#     QtGui.QMatrix4x4: 81,
#     QtGui.QVector2D: 82,
#     QtGui.QVector3D: 83,
#     QtGui.QVector4D: 84,
#     QtGui.QQuaternion: 85,
#     QtCore.QEasingCurve: 29,
#     QtCore.QJsonValue: 45,
#     QtCore.QJsonDocument: 48,
#     QtCore.QModelIndex: 42,
#     QtCore.QPersistentModelIndex: 50,
#     QtCore.QUuid: 30,
#     "user": 1024,
# }


class MetaObject:
    def __init__(self, metaobject: QtCore.QMetaObject):
        self.item = metaobject

    def __getattr__(self, val):
        return getattr(self.item, val)

    def get_super_class(self) -> MetaObject | None:
        return MetaObject(klass) if (klass := self.superClass()) is not None else None

    def get_class_info(self, include_super: bool = True) -> dict[str, str]:
        start = 0 if include_super else self.item.classInfoOffset() - 1
        count = self.item.classInfoCount()
        classinfos = [self.item.classInfo(i) for i in range(start, count)]
        return {i.name(): i.value() for i in classinfos}

    def get_method(self, index: int | str) -> core.MetaMethod:
        if isinstance(index, int):
            method = core.MetaMethod(self.item.method(index))
            if not method.isValid():
                raise KeyError(index)
            return method
        for method in self.get_methods():
            if method.get_name() == index:
                return method
        raise KeyError(index)

    def get_enum(self, index: int | str) -> core.MetaEnum:
        if isinstance(index, int):
            enum = core.MetaEnum(self.item.enumerator(index))
            if not enum.isValid():
                raise KeyError(index)
            return enum
        for enumerator in self.get_enums():
            if enumerator.get_name() == index:
                return enumerator
        raise KeyError(index)

    def get_property(self, index: int | str) -> core.MetaProperty:
        if isinstance(index, int):
            prop = core.MetaProperty(self.item.property(index))
            if not prop.isValid():
                raise KeyError(index)
            return prop
        for prop in self.get_properties():
            if prop.get_name() == index:
                return prop
        raise KeyError(index)

    def get_constructor(self, index: int | str) -> core.MetaProperty:
        if isinstance(index, int):
            method = core.MetaMethod(self.item.constructor(index))
            if not method.isValid():
                raise KeyError(index)
            return method
        for method in self.get_constructors():
            if method.get_name() == index:
                return method
        raise KeyError(index)

    def get_methods(
        self,
        include_super: bool = True,
        type_filter: core.metamethod.MethodTypeStr | None = None,
        filter_shit: bool = True,
    ) -> list[core.MetaMethod]:
        start = 0 if include_super else self.item.methodOffset() - 1
        methods = [
            method
            for i in range(start, self.item.methodCount())
            if not (method := self.get_method(i)).get_name().startswith("_q_")
            or not filter_shit
        ]
        if type_filter is None:
            return methods
        else:
            return [i for i in methods if i.get_method_type() == type_filter]

    def get_enums(self, include_super: bool = True) -> list[core.MetaEnum]:
        start = 0 if include_super else self.item.enumeratorOffset() - 1
        return [self.get_enum(i) for i in range(start, self.item.enumeratorCount())]

    def get_constructors(self) -> list[core.MetaMethod]:
        count = self.item.constructorCount()
        return [core.MetaMethod(self.item.constructor(i)) for i in range(count)]

    def get_properties(self, include_super: bool = True) -> list[core.MetaProperty]:
        start = 0 if include_super else self.item.propertyOffset() - 1
        count = self.item.propertyCount()
        return [core.MetaProperty(self.item.property(i)) for i in range(start, count)]

    def get_signals(self, include_super: bool = True) -> list[core.MetaMethod]:
        return self.get_methods(include_super=include_super, type_filter="signal")

    def get_slots(self, include_super: bool = True) -> list[core.MetaMethod]:
        return self.get_methods(include_super=include_super, type_filter="slot")

    def get_plain_methods(self, include_super: bool = True) -> list[core.MetaMethod]:
        return self.get_methods(include_super=include_super, type_filter="method")

    def get_meta_type(self) -> core.MetaType:
        return core.MetaType(self.metaType().id())

    def get_user_property(self) -> core.MetaProperty | None:
        prop = self.userProperty()
        return core.MetaProperty(prop) if prop.isValid() else None

    # just experimenting
    @classmethod
    def invoke_method(
        cls,
        obj: QtCore.QObject,
        method: str,
        *args,
        connection_type: constants.ConnectionTypeStr = "auto",
    ):
        conn = constants.CONNECTION_TYPE[connection_type]
        args = tuple(core.Q_ARG(type(arg), arg) for arg in args)
        return cls.invokeMethod(obj, method, conn, *args)

    def get_new_instance(self, *args, **kwargs):
        args = tuple(core.Q_ARG(type(i), i) for i in args)
        kwargs = {k: core.Q_ARG(type(v), v) for k, v in kwargs.items()}
        # requires QtCore.QGenericArgumentHolder for PySide6
        self.newInstance(*args, **kwargs)


if __name__ == "__main__":
    from prettyqt import widgets

    metaobj = widgets.AbstractItemView.get_static_metaobject()
    print(metaobj.get_methods())



"""Module containing helper functions."""

from __future__ import annotations

import collections
import contextlib
from collections.abc import Callable, Generator

import logging
import re
import sys

from prettyqt import gui, qt, widgets
from prettyqt.qt import QtCore
from prettyqt.utils import helpers


logger = logging.getLogger(__name__)

LOG_MAP = {
    QtCore.QtMsgType.QtDebugMsg: logging.DEBUG,
    QtCore.QtMsgType.QtInfoMsg: logging.INFO,
    QtCore.QtMsgType.QtWarningMsg: logging.WARNING,
    QtCore.QtMsgType.QtCriticalMsg: logging.ERROR,
    QtCore.QtMsgType.QtFatalMsg: logging.CRITICAL,
    QtCore.QtMsgType.QtSystemMsg: logging.CRITICAL,
}

# qFormatLogMessage(QtMsgType type, const QMessageLogContext context, const QString str)
# qInstallMessageHandler(QtMessageHandler handler)
# qSetMessagePattern(const QString &pattern)


class QtLogger(logging.Handler):
    def emit(self, record: logging.LogRecord):
        match record.level:
            case logging.DEBUG:
                QtCore.qDebug(self.format(record))
            case logging.INFO:
                QtCore.qInfo(self.format(record))
            case logging.WARNING:
                QtCore.qWarning(self.format(record))
            case logging.CRITICAL:
                QtCore.qCritical(self.format(record))
            case logging.CRITICAL:
                QtCore.qFatal(self.format(record))


class MessageHandler:
    def __init__(self, logger: logging.Logger):
        self._logger = logger
        self._previous_handler = None

    def install(self):
        self._previous_handler = QtCore.qInstallMessageHandler(self)

    def uninstall(self):
        if self._previous_handler is None:
            QtCore.qInstallMessageHandler(self._previous_handler)

    def __enter__(self):
        self.install()
        return self

    def __exit__(self, *args):
        self.uninstall()

    def __call__(
        self,
        msgtype: QtCore.QtMsgType,
        context: QtCore.QMessageLogContext,
        message: str,
    ):
        ctx = dict.fromkeys(["category", "file", "function", "line"])
        with contextlib.suppress(UnicodeDecodeError):
            ctx["category"] = context.category
        with contextlib.suppress(UnicodeDecodeError):
            ctx["file"] = context.file
        with contextlib.suppress(UnicodeDecodeError):
            ctx["function"] = context.function
        with contextlib.suppress(UnicodeDecodeError):
            ctx["line"] = context.line

        level = LOG_MAP[msgtype]
        self._logger.log(level, message, extra=ctx)


def count_objects():
    win = widgets.Application.get_mainwindow()
    objects = win.findChildren(QtCore.QObject)
    counter = collections.Counter([type(o) for o in objects])
    logger.info(counter)


def stalk(widget):
    # to activate signal logging for all
    widget.add_callback_for_event(lambda x: False)

    def make_fn(widget, signal_prop):
        def fn(*args, **kwargs):
            logger.info(f"{widget!r} {signal_prop.get_name()}")
            logger.info(f"{args} {kwargs}")

        return fn

    for signal_prop in widget.get_metaobject().get_signals():
        signal = widget.__getattribute__(signal_prop.get_name())
        fn = make_fn(widget, signal_prop)
        signal.connect(fn)


def is_deleted(obj) -> bool:
    match qt.API:
        case "pyside6":
            import shiboken6

            return not shiboken6.isValid(obj)
        case "pyqt6":
            from PyQt6 import sip

            return sip.isdeleted(obj)


class TracebackDialog(widgets.Dialog):
    """A dialog box that shows Python traceback."""

    def __init__(self, parent):
        super().__init__(parent, window_title="Traceback")
        layout = widgets.VBoxLayout()
        self.setLayout(layout)
        self._text = widgets.TextEdit(self, read_only=True, line_wrap_mode="none")
        self._text.setFontFamily(gui.Font.mono().family())
        layout.addWidget(self._text)
        self.resize(600, 400)

    def setText(self, text: str):
        """Always set text as a HTML text."""
        self._text.setHtml(text)


class ErrorMessageBox(widgets.MessageBox):
    """An message box widget for displaying Python exception."""

    def __init__(self, title: str, text_or_exception: str | Exception, parent):
        if isinstance(text_or_exception, str):
            text = text_or_exception
            exc = None
        else:
            text = str(text_or_exception)
            exc = text_or_exception
        MBox = widgets.MessageBox
        super().__init__(
            # icon=MBox.Icon.Critical,
            title=title,
            text=str(text)[:1000],
            buttons=["ok", "help", "close"],
            parent=parent,
        )

        self._exc = exc

        traceback_button = self.button(MBox.StandardButton.Help)
        traceback_button.setText("Show trackback")
        close_button = self.button(MBox.StandardButton.Close)
        close_button.setText("Quit application")

    def exec(self):
        match super().exec():
            case widgets.MessageBox.StandardButton.Help:
                tb = self._get_traceback()
                dlg = TracebackDialog(self)
                dlg.setText(tb)
                dlg.exec()
            case widgets.MessageBox.StandardButton.Close:
                sys.exit(1)
            case widgets.MessageBox.StandardButton.Ok:
                return True

    @classmethod
    def from_exc(cls, e: Exception, parent=None):
        """Construct message box from a exception."""
        return cls(type(e).__name__, e, parent)

    @classmethod
    def raise_(cls, e: Exception, parent=None):
        """Raise exception in the message box."""
        # unwrap EmitLoopError
        return cls.from_exc(e, parent=parent).exec()

    def _get_traceback(self):
        if self._exc is None:
            import traceback

            tb = traceback.format_exc()
            print(tb)
        else:
            tb = get_tb_formatter(gui.Font.mono().family())(self._exc, as_html=True)
        return tb

    @classmethod
    def _excepthook(cls, exc_type: type[Exception], exc_value: Exception, exc_traceback):
        """Exception hook used during application execution."""
        ErrorMessageBox.raise_(exc_value, parent=None)
        return None


# Following functions are mostly copied from napari (BSD 3-Clause).
# See https://github.com/napari/napari/blob/main/napari/utils/notifications.py


def get_tb_formatter(font: str = "Monospace") -> Callable[[Exception, bool, str], str]:
    """Return a formatter callable that uses IPython VerboseTB if available.

    Imports IPython lazily if available to take advantage of ultratb.VerboseTB.
    If unavailable, cgitb is used instead, but this function overrides a lot of
    the hardcoded citgb styles and adds error chaining (for exceptions that
    result from other exceptions).

    Returns:
        callable: A function that accepts a 3-tuple and a boolean ``(exc_info, as_html)``
                and returns a formatted traceback string. The ``exc_info`` tuple is of
                the ``(type, value, traceback)`` format returned by sys.exc_info().
                The ``as_html`` determines whether the traceback is formatted in html
                or plain text.
    """
    try:
        import IPython.core.ultratb

        def format_exc_info(exc: Exception, as_html: bool, color="Neutral") -> str:
            # avoids printing the array data
            # some discussion related to obtaining the current string function
            # can be found here, https://github.com/numpy/numpy/issues/11266
            info = (
                exc.__class__,
                exc,
                exc.__traceback__,
            )
            import numpy as np

            np.set_string_function(lambda arr: f"{type(arr)} {arr.shape} {arr.dtype}")

            vbtb = IPython.core.ultratb.VerboseTB(color_scheme=color)
            if as_html:
                ansi_string = (
                    vbtb.text(*info)
                    .replace(" ", "&nbsp;")
                    .replace("<", "&lt;")
                    .replace(">", "&gt;")
                )
                html = helpers.ansi2html(ansi_string)
                html = html.replace("\n", "<br>")
                html = (
                    f"<span style='font-family: monaco,{font},"
                    "monospace;'>" + html + "</span>"
                )
                tb_text = html
            else:
                tb_text = vbtb.text(*info)

            # resets to default behavior
            np.set_string_function(None)
            return tb_text

    except ModuleNotFoundError:
        import cgitb
        import traceback

        # cgitb does not support error chaining...
        # see https://peps.python.org/pep-3134/#enhanced-reporting
        # this is a workaround
        def cgitb_chain(exc: Exception) -> Generator[str, None, None]:
            """Recurse through exception stack and chain cgitb_html calls."""
            if exc.__cause__:
                yield from cgitb_chain(exc.__cause__)
                yield (
                    '<br><br><font color="#51B432">The above exception was '
                    "the direct cause of the following exception:</font><br>"
                )
            elif exc.__context__:
                yield from cgitb_chain(exc.__context__)
                yield (
                    '<br><br><font color="#51B432">During handling of the '
                    "above exception, another exception occurred:</font><br>"
                )
            yield cgitb_html(exc)

        def cgitb_html(exc: Exception) -> str:
            """Format exception with cgitb.html."""
            info = (type(exc), exc, exc.__traceback__)
            return cgitb.html(info)

        def format_exc_info(exc: Exception, as_html: bool, color=None) -> str:
            info = (
                exc.__class__,
                exc,
                exc.__traceback__,
            )
            # avoids printing the array data
            try:
                import numpy as np

                np.set_string_function(lambda arr: f"{type(arr)} {arr.shape} {arr.dtype}")
                np_imported = True
            except ImportError:
                np_imported = False
            if as_html:
                html = "\n".join(cgitb_chain(info[1]))
                # cgitb has a lot of hardcoded colors that don't work for us
                # remove bgcolor, and let theme handle it
                html = re.sub('bgcolor="#.*"', "", html)
                # remove superfluous whitespace
                html = html.replace("<br>\n", "\n")
                # but retain it around the <small> bits
                html = re.sub(r"(<tr><td><small.*</tr>)", "<br>\\1<br>", html)
                # weird 2-part syntax is a workaround for hard-to-grep text.
                html = html.replace(
                    "<p>A problem occurred in a Python script.  "
                    "Here is the sequence of",
                    "",
                )
                html = html.replace(
                    "function calls leading up to the error, "
                    "in the order they occurred.</p>",
                    "<br>",
                )
                # remove hardcoded fonts
                html = html.replace("\n", "<br>")
                html = (
                    f"<span style='font-family: monaco,{font},"
                    "monospace;'>" + html + "</span>"
                )
                tb_text = html
            else:
                # if we don't need HTML, just use traceback
                tb_text = "".join(traceback.format_exception(*info))
            # resets to default behavior
            if np_imported:
                np.set_string_function(None)
            return tb_text

    return format_exc_info


if __name__ == "__main__":
    app = widgets.app()
    widget = widgets.Widget()
    stalk(widget)
    widget.show()
    with app.debug_mode():
        app.main_loop()



from __future__ import annotations

from typing import Literal

from prettyqt import bluetooth, core
from prettyqt.qt import QtBluetooth
from prettyqt.utils import InvalidParamError, bidict, get_repr


HostMode = QtBluetooth.QBluetoothLocalDevice.HostMode
HOST_MODE = bidict(
    powered_off=HostMode.HostPoweredOff,
    connectable=HostMode.HostConnectable,
    discoverable=HostMode.HostDiscoverable,
    discoverable_limited_inquiry=HostMode.HostDiscoverableLimitedInquiry,
)

HostModeStr = Literal[
    "powered_off", "connectable", "discoverable", "discoverable_limited_inquiry"
]

Error = QtBluetooth.QBluetoothLocalDevice.Error

ERROR = bidict(
    none=Error.NoError,
    pairing=Error.PairingError,
    missing_permissions=Error.MissingPermissionsError,
    unknown=Error.UnknownError,
)

ErrorStr = Literal["none", "pairing", "missing_permissions", "unknown"]


Pairing = QtBluetooth.QBluetoothLocalDevice.Pairing

PAIRING = bidict(
    unpaired=Pairing.Unpaired,
    paired=Pairing.Paired,
    authorized_paired=Pairing.AuthorizedPaired,
)

PairingStr = Literal["unpaired", "paired", "authorized_paired"]


class BluetoothLocalDevice(core.ObjectMixin, QtBluetooth.QBluetoothLocalDevice):
    def __bool__(self):
        return self.isValid()

    def __repr__(self):
        return get_repr(self, self.address())

    def get_error(self) -> ErrorStr:
        return ERROR.inverse[self.error()]

    def set_host_mode(self, mode: HostModeStr):
        """Set host mode.

        Args:
            mode: host mode to use

        Raises:
            InvalidParamError: invalid host mode
        """
        if mode not in HOST_MODE:
            raise InvalidParamError(mode, HOST_MODE)
        self.setGridStyle(HOST_MODE[mode])

    def get_host_mode(self) -> HostModeStr:
        """Return host mode.

        Returns:
            host mode
        """
        return HOST_MODE.inverse[self.gridStyle()]

    def get_pairing_status(
        self, address: QtBluetooth.QBluetoothAddress | int | str
    ) -> PairingStr:
        """Return pairing status.

        Arguments:
            address: bluetooth address

        Returns:
            pairing status
        """
        if isinstance(address, int | str):
            address = QtBluetooth.QBluetoothAddress(address)
        return PAIRING.inverse[self.pairingStatus(address)]

    def request_pairing(
        self, address: QtBluetooth.QBluetoothAddress | int | str, pairing: PairingStr
    ):
        if isinstance(address, int | str):
            address = QtBluetooth.QBluetoothAddress(address)
        self.requestPairing(address, PAIRING[pairing])

    def get_connected_devices(self) -> list[bluetooth.BluetoothAddress]:
        return [bluetooth.BluetoothAddress(i) for i in self.connectedDevices()]

    @classmethod
    def get_all_devices(cls) -> list[bluetooth.BluetoothHostInfo]:
        return [bluetooth.BluetoothHostInfo(i) for i in cls.allDevices()]


if __name__ == "__main__":
    devices = BluetoothLocalDevice.get_all_devices()
    print(devices)



from __future__ import annotations

from prettyqt import constants, core
from prettyqt.qt import QtWidgets


class AbstractItemDelegateMixin(core.ObjectMixin):
    def _data_for_index(self, index, role=constants.USER_ROLE):
        # using index.data() sometimes casts stuff in PyQt6
        model = index.model()
        data = model.data(index, role)
        return data


class AbstractItemDelegate(AbstractItemDelegateMixin, QtWidgets.QAbstractItemDelegate):
    pass



from __future__ import annotations

from prettyqt import core, widgets


class CollapsibleFrame(widgets.Frame):
    expanded = core.Signal()
    collapsed = core.Signal()

    def __init__(self, text: str = "", **kwargs):
        super().__init__(frame_shape="styled_panel", frame_shadow="plain", **kwargs)
        _layout = widgets.VBoxLayout()
        _layout.set_margin(0)
        _layout.setSpacing(0)
        self.setLayout(_layout)
        # button
        self._button = widgets.ToolButton(
            self,
            clicked=self.on_button_click,
            arrow_type="right",
            auto_raise=False,
            text=text,
            visible=True,
        )
        self._button.set_style("text_beside_icon")
        self.set_size_policy("minimum_expanding", "fixed")
        _layout.addWidget(self._button, 0)
        # group box
        self._panel = widgets.Widget(self, visible=False)
        _layout.addWidget(self._panel)
        self._panel_layout = widgets.VBoxLayout()
        self._panel_layout.set_margin(1)
        self._panel_layout.setSpacing(2)
        self._panel.setLayout(self._panel_layout)
        # connect signals
        # private state variables
        self._is_collapsed = True

    def set_title(self, title: str):
        self._button.set_text(title)

    def add_widget(self, widget: widgets.Widget):
        self._panel_layout.addWidget(widget)

    def remove_widget(self, widget: widgets.Widget):
        self._panel_layout.removeWidget(widget)

    def is_expanded(self) -> bool:
        return not self._is_collapsed

    def expand(self):
        self._button.set_arrow_type("down")
        self._panel.setVisible(True)
        self._is_collapsed = False
        self.set_size_policy("minimum_expanding", "minimum_expanding")

    def collapse(self):
        self._panel.setVisible(False)
        self._button.set_arrow_type("right")
        self._is_collapsed = True
        self.set_size_policy("preferred", "preferred")

    @core.Slot()
    def on_button_click(self):
        if self._is_collapsed:
            self.expand()
            self.expanded.emit()
        else:
            self.collapse()
            self.collapsed.emit()


if __name__ == "__main__":
    app = widgets.app()
    widget = CollapsibleFrame()
    widget.add_widget(widgets.Label("test"))
    widget.show()
    app.main_loop()



from __future__ import annotations

from collections.abc import Iterator
from typing import Literal

from deprecated import deprecated
from typing_extensions import Self

from prettyqt import constants, core, widgets
from prettyqt.qt import QtCore, QtWidgets
from prettyqt.utils import InvalidParamError, bidict, get_repr, prettyprinter


SIZE_CONSTRAINT = bidict(
    default=QtWidgets.QLayout.SizeConstraint.SetDefaultConstraint,
    fixed=QtWidgets.QLayout.SizeConstraint.SetFixedSize,
    minimum=QtWidgets.QLayout.SizeConstraint.SetMinimumSize,
    maximum=QtWidgets.QLayout.SizeConstraint.SetMaximumSize,
    min_and_max=QtWidgets.QLayout.SizeConstraint.SetMinAndMaxSize,
    none=QtWidgets.QLayout.SizeConstraint.SetNoConstraint,
)

SizeConstraintStr = Literal[
    "default", "fixed", "minimum", "maximum", "min_and_max", "none"
]

LayoutTypeStr = Literal["horizontal", "vertical", "grid", "form", "stacked", "flow"]


class LayoutMixin(core.ObjectMixin, widgets.LayoutItemMixin, prettyprinter.PrettyPrinter):
    def __init__(self, *args, margin=None, **kwargs):
        self._next_container = None
        self._stack = []
        super().__init__(*args, **kwargs)
        if margin is not None:
            self.set_margin(margin)

    def __getitem__(
        self, index: str | int
    ) -> QtWidgets.QWidget | QtWidgets.QLayout | None:
        if isinstance(index, int):
            item = self.itemAt(index)
            widget = item.widget()
            if widget is None:
                widget = item.layout()
        elif isinstance(index, str):
            return self.find_child(typ=QtCore.QObject, name=index)
        return widget

    def __delitem__(self, index: int):
        item = self.itemAt(index)
        self.removeItem(item)

    def __len__(self) -> int:
        return self.count()

    def __repr__(self):
        return get_repr(self)

    def __iter__(self) -> Iterator[QtWidgets.QWidget | QtWidgets.QLayout | None]:
        return iter(self[i] for i in range(self.count()))

    def __contains__(self, item: QtWidgets.QWidget | QtWidgets.QLayoutItem):
        return self.indexOf(item) >= 0

    def __enter__(self):
        if self._next_container is not None:
            self._next_container.__enter__()
            self._stack.append(self._next_container)
            self._next_container = None
        return self

    def __exit__(self, *_):
        if self._stack:
            item = self._stack.pop()
            item.__exit__()

    def __iadd__(self, item, *args, **kwargs):
        self.add(item, *args, **kwargs)
        return self

    # def __getattr__(self, name):
    #     return getattr(self._layout, name)

    # def __call__(self):
    #     return self._layout

    def _get_map(self):
        maps = super()._get_map()
        maps |= {"sizeConstraint": SIZE_CONSTRAINT}
        return maps

    def add(self, item, *args, **kwargs):
        if isinstance(item, QtWidgets.QWidget):
            self._layout.addWidget(item, *args, **kwargs)
        elif isinstance(item, QtWidgets.QLayout):
            self._layout.addLayout(item, *args, **kwargs)
            # widget = Widgets.Widget()
            # widget.set_layout(item)
            # self._layout.addWidget(item, *args, **kwargs)
        elif isinstance(item, list):
            for i in item:
                self._layout.add(i, *args, **kwargs)

        return item

    @property
    def _layout(self):
        return self._stack[-1] if self._stack else self

    @classmethod
    def create(
        cls,
        parent: QtWidgets.QWidget | QtWidgets.QLayout | None = None,
        stretch: int | None = None,
        margin: int | None = None,
        align: constants.AlignmentStr | None = None,
        **kwargs,
    ) -> Self:
        """Create a Layout attached to given parent. Insert widget if needed."""
        match parent:
            case QtWidgets.QMainWindow():
                widget = widgets.Widget(parent=parent)
                parent.setCentralWidget(widget)
                new = cls(widget, **kwargs)
            case QtWidgets.QScrollArea():
                widget = widgets.Widget(parent=parent)
                parent.setWidget(widget)
                new = cls(widget, **kwargs)
            case QtWidgets.QSplitter():
                widget = widgets.Widget(parent=parent)
                parent.addWidget(widget)
                new = cls(widget, **kwargs)
            case None | QtWidgets.QWidget():
                new = cls(parent, **kwargs)
            case QtWidgets.QLayout():
                new = cls(**kwargs)
                if stretch:
                    parent.addLayout(new, stretch)
                else:
                    parent.addLayout(new)

        if margin is not None:
            new.set_margin(margin)
        if align is not None:
            new.set_alignment(align)

        new._stack = []
        new._next_container = None
        return new

    def get_sub_layout(self, layout: str, *args, **kwargs) -> Self:
        match layout:
            case "horizontal":
                Class = widgets.HBoxLayout
            case "vertical":
                Class = widgets.VBoxLayout
            case "grid":
                Class = widgets.GridLayout
            case "form":
                Class = widgets.FormLayout
            case "stacked":
                Class = widgets.StackedLayout
            case "flow":
                from prettyqt import custom_widgets

                Class = custom_widgets.FlowLayout
            case "splitter":
                Class = widgets.Splitter
            case "scroll":
                Class = widgets.ScrollArea
            case _:
                raise ValueError("Invalid Layout")
        self._next_container = Class.create(self._layout, *args, **kwargs)
        return self

    def get_children(self) -> list[QtWidgets.QWidget | QtWidgets.QLayout]:
        return list(self)

    def set_margin(self, margin: tuple[int, int, int, int] | int | None):
        match margin:
            case None:
                self.unsetContentsMargins()
            case int():
                self.setContentsMargins(margin, margin, margin, margin)
            case tuple():
                self.setContentsMargins(*margin)
            case _:
                raise ValueError(margin)

    def set_spacing(self, pixels: int):
        self.setSpacing(pixels)

    @deprecated(reason="Use set_size_constraint instead")
    def set_size_mode(self, mode: SizeConstraintStr):
        self.set_size_constraint(mode)

    def set_size_constraint(self, mode: SizeConstraintStr):
        """Set the size mode of the layout.

        Args:
            mode: size mode for the layout

        Raises:
            InvalidParamError: size mode does not exist
        """
        if mode not in SIZE_CONSTRAINT:
            raise InvalidParamError(mode, SIZE_CONSTRAINT)
        self.setSizeConstraint(SIZE_CONSTRAINT[mode])

    @deprecated(reason="Use set_size_constraint instead")
    def get_size_mode(self) -> SizeConstraintStr:
        return self.get_size_constraint()

    def get_size_constraint(self) -> SizeConstraintStr:
        """Return current size mode.

        Returns:
            size mode
        """
        return SIZE_CONSTRAINT.inverse[self.sizeConstraint()]

    def set_alignment(
        self,
        alignment: constants.AlignmentStr,
        item: QtWidgets.QWidget | QtWidgets.QLayout | None = None,
    ) -> bool:
        """Set the alignment for widget / layout to alignment.

        Returns true if w is found in this layout (not including child layouts).

        Args:
            alignment: alignment for the layout
            item: set alignment for specific child only

        Raises:
            InvalidParamError: alignment does not exist
        """
        if alignment not in constants.ALIGNMENTS:
            raise InvalidParamError(alignment, constants.ALIGNMENTS)
        if item is not None:
            return self.setAlignment(item, constants.ALIGNMENTS[alignment])
        else:
            return self.setAlignment(constants.ALIGNMENTS[alignment])

    def add_widget(self, widget: QtWidgets.QWidget, *args, **kwargs):
        self.addWidget(widget, *args, **kwargs)

    # def add(self, *items: QtWidgets.QWidget | QtWidgets.QLayout):
    #     for i in items:
    #         match i:
    #             case QtWidgets.QWidget():
    #                 self.addWidget(i)
    #             case QtWidgets.QLayout():
    #                 w = widgets.Widget()
    #                 w.set_layout(i)
    #                 self.addWidget(w)
    #             case _:
    #                 raise TypeError("add_item only supports widgets and layouts")


class Layout(LayoutMixin, QtWidgets.QLayout):
    pass


if __name__ == "__main__":
    app = widgets.app()
    widget = widgets.ScrollArea()

    with widgets.VBoxLayout.create(widget) as layout:
        with layout.get_sub_layout("splitter", orientation="horizontal") as layout:
            layout += widgets.PlainTextEdit("upper left")
            layout += widgets.PlainTextEdit("upper middle")
            with layout.get_sub_layout("splitter", orientation="vertical") as layout:
                layout += widgets.PlainTextEdit("upper right")
                layout += widgets.PlainTextEdit("middle right")
                with layout.get_sub_layout("horizontal") as layout:
                    layout += widgets.PlainTextEdit("upper right")
                    layout += widgets.PlainTextEdit("middle right")
        with layout.get_sub_layout("horizontal") as layout:
            layout += widgets.PlainTextEdit("lower left")
            layout += widgets.PlainTextEdit("lower right")
    widget.setWidgetResizable(True)

    widget.show()
    app.main_loop()



from __future__ import annotations

from typing import Literal

from prettyqt import widgets
from prettyqt.custom_widgets.frameless.windows import utils
from prettyqt.qt import QtCore, QtGui, QtWidgets


class TitleBarButton(widgets.PushButton):
    def __init__(self, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)
        self.setFixedSize(48, 32)
        self.set_icon_size((28, 28))


class MaximizeButton(TitleBarButton):
    def __init__(self, parent: QtWidgets.QWidget | None) -> None:
        super().__init__(parent)
        self.setStyleSheet(
            """
        QPushButton {
            background-position: center;
            background-repeat: no-repeat;
            border: none;
            outline: none;
            color: #d1d3d2;
        }

        QPushButton::hover {
            background-color: #1D1D24;
        }
        """
        )

        self.set_icon("mdi.window-maximize")

    def set_state(self, state: Literal["hover", "normal"]) -> None:
        if state == "hover":
            self.setStyleSheet(
                """
                    QPushButton {
                        background-position: center;
                        background-repeat: no-repeat;
                        border: none;
                        outline: none;
                        color: #d1d3d2;
                        background-color: #1D1D24;
                    }
                    """
            )
        elif state == "normal":
            self.setStyleSheet(
                """
                    QPushButton {
                        background-position: center;
                        background-repeat: no-repeat;
                        border: none;
                        outline: none;
                        color: #d1d3d2;
                        background: transparent;
                    }
                    """
            )
        self.update()


class MinimizeButton(TitleBarButton):
    def __init__(self, parent: QtWidgets.QWidget | None) -> None:
        super().__init__(parent)
        self.setStyleSheet(
            """
                QPushButton {
            background-position: center;
            background-repeat: no-repeat;
            border: none;
            outline: none;
            color: #d1d3d2;
        }

        QPushButton::hover {
            background-color: #1D1D24;
        }
        """
        )
        self.set_icon("mdi.window-minimize")


class CloseButton(TitleBarButton):
    def __init__(self, parent: QtWidgets.QWidget | None) -> None:
        super().__init__(parent)
        self.setStyleSheet(
            """
        QPushButton {
            background-position: center;
            background-repeat: no-repeat;
            border: none;
            outline: none;
            color: #d1d3d2;
        }

        QPushButton:hover {
            background-color: #e81123;
        }"""
        )
        self.set_icon("mdi.window-close")


class TitleBar(widgets.Frame):
    def __init__(self, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)
        self.setFixedHeight(32)

        self.button_box = widgets.Widget(self)

        self.maximize_button = MaximizeButton(self.button_box)
        self.minimize_button = MinimizeButton(self.button_box)
        self.close_button = CloseButton(self.button_box)

        self.bbox_layout = widgets.HBoxLayout(parent=self.button_box)
        self.bbox_layout.set_margin(0)
        self.bbox_layout.setSpacing(0)
        for btn in [self.minimize_button, self.maximize_button, self.close_button]:
            self.bbox_layout.addWidget(btn)

        self.horizontal_layout = widgets.HBoxLayout(parent=self)
        self.horizontal_layout.set_margin(0)
        self.horizontal_layout.setSpacing(0)
        self.horizontal_spacer = widgets.SpacerItem(20, 20, "expanding", "minimum")
        self.horizontal_layout.addSpacerItem(self.horizontal_spacer)
        self.horizontal_layout.addWidget(self.button_box)
        self.minimize_button.clicked.connect(self.on_minimize_button_clicked)
        self.maximize_button.clicked.connect(self.on_maximize_button_clicked)
        self.close_button.clicked.connect(self.on_close_button_clicked)

        self.window().installEventFilter(self)

    def on_close_button_clicked(self) -> None:
        self.window().close()

    def on_maximize_button_clicked(self):
        if self.window().isMaximized():
            self.window().showNormal()
            self.set_maximize_button_icon("maximize")
        else:
            self.window().showMaximized()
            self.set_maximize_button_icon("restore")

    def on_minimize_button_clicked(self) -> None:
        self.window().showMinimized()

    def set_maximize_button_icon(self, icon: Literal["maximize", "restore"]) -> None:
        if icon == "maximize":
            self.maximize_button.set_icon("mdi.window-maximize")
        elif icon == "restore":
            self.maximize_button.set_icon("mdi.window-restore")

    def eventFilter(self, obj, e) -> bool:
        if obj is self.window() and e.type() == QtCore.QEvent.Type.WindowStateChange:
            if self.window().isMaximized():
                self.set_maximize_button_icon("restore")

            else:
                self.set_maximize_button_icon("maximize")
        return super().eventFilter(obj, e)

    def mouseMoveEvent(self, event: QtGui.QMouseEvent) -> None:
        utils.start_system_move(self.window().winId())


class FramelessMixin:
    def __init__(self, parent: QtWidgets.QWidget | None = None, **kwargs) -> None:
        super().__init__(parent, **kwargs)
        self.setWindowFlags(
            QtCore.Qt.WindowType.Window | QtCore.Qt.WindowType.FramelessWindowHint
        )
        self.title_bar = TitleBar(self)
        win_id = self.winId()
        utils.add_shadow_effect(win_id)
        utils.add_window_animation(win_id)

        self.resize(800, 800)

    def set_nonresizable(self):
        utils.set_window_nonresizable(self.winId())
        self.title_bar.maximize_button.hide()

    def is_resizable(self) -> None:
        return utils.is_window_resizable(self.winId())

    def showEvent(self, event) -> None:
        self.title_bar.raise_()
        super().showEvent(event)

    def nativeEvent(self, event_type: QtCore.QByteArray, message: int):
        ret_tuple = utils._native_event(self, message)
        if ret_tuple is not None:
            ret, value = ret_tuple
            if ret:
                return ret, value
        super().nativeEvent(event_type, message)

    def resizeEvent(self, e):
        super().resizeEvent(e)
        self.title_bar.resize(self.width(), self.title_bar.height())


if __name__ == "__main__":
    app = widgets.app()
    app.set_style("Fusion")

    class MainWindow(FramelessMixin, widgets.MainWindow):
        pass

    m = MainWindow()
    button = widgets.PushButton("test")
    m.setCentralWidget(button)
    m.show()
    m.resize(240, 160)
    app.main_loop()



from __future__ import annotations

from datetime import timedelta
import re


TIME_REGEX = re.compile(
    r"((?P<hours>\d+?)h)?((?P<minutes>\d+?)m)?"
    r"((?P<seconds>\d+?)s)?((?P<milliseconds>\d+?)ms)?"
)

CASE_PATTERN = re.compile(r"(?<!^)(?=[A-Z])")


def dump_json(data: str):
    try:
        import orjson

        opts = orjson.OPT_NAIVE_UTC | orjson.OPT_SERIALIZE_NUMPY
        return orjson.dumps(data, option=opts)
    except ImportError:
        import json

        return json.dumps(data).encode()


def load_json(data):
    try:
        import orjson

        return orjson.loads(data)
    except ImportError:
        import json

        return json.loads(data)


def parse_time(time_str: str) -> int:
    parts = TIME_REGEX.match(time_str)
    if not parts:
        raise ValueError(time_str)
    dct = parts.groupdict()
    time_params = {name: int(param) for (name, param) in dct.items() if param}
    secs = timedelta(**time_params).total_seconds()
    return int(secs * 1000)


def to_lower_camel(snake_str: str) -> str:
    # do nothing if nothing to camel
    if "_" not in snake_str:
        return snake_str
    first, *others = snake_str.split("_")
    return "".join([first.lower(), *map(str.title, others)])


def to_snake(camel_string):
    #  don´t snake-case snakes.
    if "_" in camel_string:
        return camel_string
    return CASE_PATTERN.sub("_", camel_string).lower()


def string_to_num_array(array: str) -> list[float]:
    floats = [float(i) for i in array.split(",")]
    return [int(i) if i.is_integer() else i for i in floats]


def get_color_percentage(
    color_1: tuple[int, int, int, int], color_2: tuple[int, int, int, int], percent: float
) -> tuple[int, int, int, int]:
    """Get a color which is percent% interpolated between start and end.

    Args:
        color_1 : Start color components (R, G, B, A / H, S, V, A / H, S, L, A)
        color_2 : End color components (R, G, B, A / H, S, V, A / H, S, L, A)
        percent: Percentage to interpolate, 0-100.
                 0: Start color will be returned.
                 100: End color will be returned.

    Return:
        A (x, y, z, alpha) tuple with the interpolated color components.
    """
    if not 0 <= percent <= 100:
        raise ValueError("percent needs to be between 0 and 100!")
    x = round(color_1[0] + (color_2[0] - color_1[0]) * percent / 100)
    y = round(color_1[1] + (color_2[1] - color_1[1]) * percent / 100)
    z = round(color_1[2] + (color_2[2] - color_1[2]) * percent / 100)
    a = round(color_1[3] + (color_2[3] - color_1[3]) * percent / 100)
    return (x, y, z, a)


ANSI_STYLES = {
    1: {"font_weight": "bold"},
    2: {"font_weight": "lighter"},
    3: {"font_weight": "italic"},
    4: {"text_decoration": "underline"},
    5: {"text_decoration": "blink"},
    6: {"text_decoration": "blink"},
    8: {"visibility": "hidden"},
    9: {"text_decoration": "line-through"},
    30: {"color": "black"},
    31: {"color": "red"},
    32: {"color": "green"},
    33: {"color": "yellow"},
    34: {"color": "blue"},
    35: {"color": "magenta"},
    36: {"color": "cyan"},
    37: {"color": "white"},
}


def ansi2html(ansi_string: str, styles: dict[int, dict[str, str]] = ANSI_STYLES) -> str:
    """Convert ansi string to colored HTML.

    Arguments:
        ansi_string:
            text with ANSI color codes.
        styles:
            A mapping from ANSI codes to a dict of css kwargs:values,
            by default ANSI_STYLES

    Returns:
        HTML string
    """
    previous_end = 0
    in_span = False
    ansi_codes = []
    ansi_finder = re.compile("\033\\[([\\d;]*)([a-zA-z])")
    parts = []
    for match in ansi_finder.finditer(ansi_string):
        parts.append(ansi_string[previous_end : match.start()])
        previous_end = match.end()
        params, command = match.groups()

        if command not in "mM":
            continue

        try:
            params = [int(p) for p in params.split(";")]
        except ValueError:
            params = [0]

        for i, v in enumerate(params):
            if v == 0:
                params = params[i + 1 :]
                if in_span:
                    in_span = False
                    parts.append("</span>")
                ansi_codes = []
                if not params:
                    continue

        ansi_codes.extend(params)
        if in_span:
            parts.append("</span>")
            in_span = False

        if not ansi_codes:
            continue

        style = [
            "; ".join([f"{k}: {v}" for k, v in styles[k].items()]).strip()
            for k in ansi_codes
            if k in styles
        ]
        parts.append(f'<span style="{"; ".join(style)}">')

        in_span = True

    parts.append(ansi_string[previous_end:])
    if in_span:
        parts.append("</span>")
        in_span = False
    return "".join(parts)



from __future__ import annotations

import enum
import logging

from prettyqt import constants, core, widgets
from prettyqt.utils import fuzzy
from prettyqt.qt import QtCore

logger = logging.getLogger(__name__)


class FuzzyFilterProxyModel(core.SortFilterProxyModel):
    """Proxy model with fuzzyfilter functionality.

    this proxymodel replaces the text from the display role in the given filter column
    with HTML code in order to color the letter matches. A backup from the original text
    is made available in the BackupRole. Based on the original text, the proxy calculates
    a score for the match and makes it available via the SortRole.
    To display the html code properly, a HtmlItemDelegate is needed.

    """

    class Roles(enum.IntEnum):
        """Addional roles."""

        BackupRole = constants.USER_ROLE + 65
        SortRole = constants.SORT_ROLE

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._search_term = ""
        self.match_color: str | None = "blue"
        self.setSortRole(self.Roles.SortRole)
        self.sort(0, constants.DESCENDING)

    def set_match_color(self, color):
        self.match_color = color

    def filterAcceptsRow(self, source_row: int, source_index: core.ModelIndex) -> bool:
        if self._search_term == "":
            return True
        column = self.filterKeyColumn()
        source_model = self.sourceModel()
        idx = source_model.index(source_row, column, source_index)
        text = source_model.data(idx)
        return fuzzy.fuzzy_match_simple(
            self._search_term, text, case_sensitive=self.is_filter_case_sensitive()
        )

    def lessThan(self, left, right):
        role = super().sortRole()
        left_data = left.data(role)
        right_data = right.data(role)
        if left_data is None or right_data is None:
            return True
        if self._search_term:
            return fuzzy.fuzzy_match(self._search_term, left_data) < fuzzy.fuzzy_match(
                self._search_term, right_data
            )
        else:
            return left_data < right_data

    def set_search_term(self, search_term: str):
        self._search_term = search_term
        self.invalidate()

    def data(self, index, role=constants.DISPLAY_ROLE):
        if not index.isValid():
            return None
        filter_column = self.filterKeyColumn()
        match role, index.column():
            case constants.DISPLAY_ROLE, _ if index.column() == filter_column:
                label = super().data(index, constants.DISPLAY_ROLE)
                # logging.info(label)
                return (
                    fuzzy.color_text(
                        self._search_term,
                        str(label),
                        self.match_color,
                        self.is_filter_case_sensitive(),
                    )
                    if self._search_term and self.match_color
                    else label
                )
            # case constants.DISPLAY_ROLE, 1:
            #     idx = self.index(index.row(), filter_column)
            #     label = super().data(idx, constants.DISPLAY_ROLE)
            #     result = fuzzy.fuzzy_match(self._search_term, label)
            #     return str(result[1])
            case self.Roles.BackupRole, _:
                return super().data(index, constants.DISPLAY_ROLE)
            case self.Roles.SortRole, _:
                idx = self.index(index.row(), filter_column)
                label = super().data(idx, constants.DISPLAY_ROLE)
                result = fuzzy.fuzzy_match(self._search_term, label)
                return result[1]
            case _, _:
                return super().data(index, role)


class FuzzyCompleter(widgets.Completer):
    def __init__(self, parent):
        super().__init__(parent)
        parent.setEditable(True)
        parent.set_insert_policy("no_insert")
        parent.installEventFilter(self)
        self.set_completion_mode("popup")
        self._local_completion_prefix = ""
        self._source_model = None
        self._filter_proxy = FuzzyFilterProxyModel(self)
        self._filter_proxy.set_match_color(None)
        self._using_original_model = False

    def setModel(self, model):
        self._source_model = model
        self._filter_proxy = FuzzyFilterProxyModel(self)
        self._filter_proxy.set_match_color(None)
        self._filter_proxy.setSourceModel(self._source_model)
        super().setModel(self._filter_proxy)
        self._using_original_model = True

    def eventFilter(self, source: QtCore.QObject, event: QtCore.QEvent) -> bool:
        match event.type():
            case QtCore.QEvent.Type.FocusIn:
                source.clearEditText()
            case QtCore.QEvent.Type.KeyPress:
                key = event.key()
                if key == QtCore.Qt.Key.Key_Enter:
                    text = source.currentText()
                    source.setCompleter(None)
                    source.setEditText(text)
                    source.setCompleter(source.comp)
        return super().eventFilter(source, event)

    def updateModel(self):
        if not self._using_original_model:
            self._filter_proxy.setSourceModel(self._source_model)

        self._filter_proxy.set_search_term(self._local_completion_prefix)

    def splitPath(self, path):
        self._local_completion_prefix = path
        self.updateModel()
        if self._filter_proxy.rowCount() == 0:
            self._using_original_model = False
            self._filter_proxy.setSourceModel(core.StringListModel([path]))
            return [path]

        return []


if __name__ == "__main__":
    from prettyqt import custom_delegates, widgets
    from prettyqt.custom_models import JsonModel

    app = widgets.app()
    dist = [
        dict(
            assss=2,
            bffff={
                "a": 4,
                "b": [1, 2, 3],
                "jkjkjk": "tekjk",
                "sggg": "tekjk",
                "fdfdf": "tekjk",
                "xxxx": "xxx",
            },
        ),
        6,
        "jkjk",
    ]

    _source_model = JsonModel(dist)
    model = FuzzyFilterProxyModel()
    model.setFilterKeyColumn(1)
    cb = widgets.ComboBox()
    completer = FuzzyCompleter(cb)
    cb.setCompleter(completer)  #
    completer.set_strings(["Lola", "Lila", "Cola", "Lothian"])
    # cb.setModel(model)
    # cb.comp.setModel(model)
    # model.set_search_term("tj")
    model.setSourceModel(_source_model)
    widget = widgets.Widget()
    widget.set_layout("vertical")
    lineedit = widgets.LineEdit()
    lineedit.value_changed.connect(model.set_search_term)
    delegate = custom_delegates.HtmlItemDelegate()
    table = widgets.TreeView()
    table.setItemDelegateForColumn(1, delegate)
    table.setRootIsDecorated(True)
    # table.setSortingEnabled(True)
    table.set_model(model)
    widget.box.add(cb)
    widget.box.add(lineedit)
    widget.box.add(table)
    widget.show()
    with app.debug_mode():
        app.main_loop()



from __future__ import annotations

from collections.abc import Iterable
from dataclasses import dataclass
import datetime
import enum
import itertools
import os
import pathlib
from typing import TypedDict

import fsspec

from prettyqt import constants, core, custom_models, widgets
from prettyqt.qt import QtCore
from prettyqt.utils import treeitem


class FsSpecCompleter(widgets.Completer):
    def pathFromIndex(self, index):
        return index.data(self.model().Roles.FilePathRole)

    def splitPath(self, path: str):
        return path.split(self.model().fs.sep)


class FolderInfo(TypedDict):
    name: str
    size: int
    type: str
    created: float
    islink: bool  # symbolic link
    mode: int  # Inode protection mode.
    uid: int  # user id of owner
    gid: int  # group id of owner
    mtime: float
    ino: int  # Inode number.
    nlink: int  # Number of links to the inode.


_icon_provider = widgets.FileIconProvider()


VALUE_LETTERS = [(4, "r"), (2, "w"), (1, "x")]


def octal_to_string(octal: int) -> str:
    result = ""
    # Iterate over each of the digits in octal
    for digit, (value, letter) in itertools.product(
        [int(n) for n in str(octal)], VALUE_LETTERS
    ):
        if digit >= value:
            result += letter
            digit -= value
        else:
            result += "-"
    return result


def get_filename(path):
    return pathlib.Path(path).name if path else ""


@dataclass  # (frozen=True)
class FsSpecColumnItem(custom_models.ColumnItem):
    identifier: str = ""


loc = core.Locale()


ATTR_MODEL_NAME = FsSpecColumnItem(
    identifier="name",
    name="Name",
    doc="The name of the object.",
    label=lambda x: get_filename(x.obj["name"]),
    set_edit=lambda x, value: x.set_name(value),
    decoration=lambda x: _icon_provider.get_icon(core.FileInfo(x.obj["name"])),
    user_data={
        constants.USER_ROLE + 1: lambda x: x.obj["name"],
        constants.USER_ROLE + 2: lambda x: get_filename(x.obj["name"]),
    },
)


ATTR_MODEL_PATH = FsSpecColumnItem(
    identifier="name",
    name="Path",
    doc="A path",
    label=lambda x: x.obj["name"] or "",
)

ATTR_MODEL_SIZE = FsSpecColumnItem(
    identifier="size",
    name="Size",
    doc="Item size.",
    label=lambda x: loc.get_formatted_data_size(x.obj["size"])
    if x.obj["size"] > 0
    else "",
    sort_value=lambda x: x.obj["size"],
)

ATTR_MODEL_TYPE = FsSpecColumnItem(
    identifier="type",
    name="Type",
    doc="Item type.",
    label=lambda x: x.obj["type"] or "",
)

ATTR_MODEL_CREATED = FsSpecColumnItem(
    identifier="created",
    name="Created",
    doc="Date created.",
    label=lambda x: datetime.datetime.fromtimestamp(x.obj["created"])
    if x.obj.get("created")
    else "",
)

ATTR_MODEL_MODIFIED = FsSpecColumnItem(
    identifier="mtime",
    name="Modified",
    doc="Date modified.",
    label=lambda x: datetime.datetime.fromtimestamp(x.obj["mtime"])
    if x.obj.get("mtime")
    else "",
)

ATTR_MODEL_PERMISSIONS = FsSpecColumnItem(
    identifier="mode",
    name="Permissions",
    doc="File permissions.",
    label=lambda x: oct(int(x.obj["mode"]))[-4:] if x.obj.get("mode") else "",
)

ATTR_MODEL_IS_LINK = FsSpecColumnItem(
    identifier="islink",
    name="Link",
    doc="Symbolic link.",
    checkstate=lambda x: x.obj.get("islink") or False,
)

ATTR_MODEL_SHA = FsSpecColumnItem(
    identifier="sha",
    name="SHA",
    doc="SHA",
    label=lambda x: x.obj.get("sha") or "",
)

COLUMNS = [
    ATTR_MODEL_NAME,
    ATTR_MODEL_PATH,
    ATTR_MODEL_TYPE,
    ATTR_MODEL_MODIFIED,
    ATTR_MODEL_SIZE,
    ATTR_MODEL_CREATED,
    ATTR_MODEL_PERMISSIONS,
    ATTR_MODEL_IS_LINK,
    ATTR_MODEL_SHA,
]


class FSSpecTreeModel(
    widgets.filesystemmodel.FileSystemModelMixin, custom_models.ColumnItemModel
):
    """Model that provides an interface to an objectree that is build of tree items."""

    directoryLoaded = core.Signal(str)
    fileRenamed = core.Signal(str, str, str)
    rootPathChanged = core.Signal(str)

    class Roles(enum.IntEnum):
        """Role enum."""

        FileIconRole = constants.DECORATION_ROLE
        FilePathRole = constants.USER_ROLE + 1
        FileNameRole = constants.USER_ROLE + 2
        FilePermissions = constants.USER_ROLE + 3

    class Option(enum.IntEnum):
        """Role enum."""

        DontWatchForChanges = 1
        DontResolveSymlinks = 2
        DontUseCustomDirectoryIcons = 4

    def __init__(
        self,
        protocol: str = "file",
        root: os.PathLike = "",
        show_root: bool = False,
        parent: QtCore.QObject | None = None,
        **kwargs,
    ):
        self.set_protocol(protocol, **kwargs)
        obj = self.fs.info(root)
        columns = self._get_columns_for_protocol(protocol)
        super().__init__(obj, columns=columns, parent=parent, show_root=show_root)

    def set_protocol(self, protocol: str, **kwargs):
        self.fs = fsspec.filesystem(protocol, **kwargs)

    def _get_columns_for_protocol(self, protocol: str):
        match self.fs.protocol:
            case "github":
                return [
                    col
                    for col in COLUMNS
                    if col.identifier in {"name", "mode", "type", "size", "sha"}
                ]
            case "file":
                return [col for col in COLUMNS if col.identifier != "sha"]
            case _:
                return COLUMNS

    def _fetch_object_children(self, obj: treeitem.TreeItem) -> list[treeitem.TreeItem]:
        """Fetch the children of a Python object.

        Returns: list of TreeItems
        """
        glob = f"{obj.obj['name']}/*/" if obj.obj["name"] else "*"
        items = [
            treeitem.TreeItem(obj=i, parent=obj)
            for i in self.fs.glob(glob, detail=True).values()
        ]
        # not sure if this should be emitted later?
        self.directoryLoaded.emit(obj.obj["name"])
        return items

    def get_protocol_path(self, index: QtCore.QModelIndex) -> str:
        protocol = self.fs.protocol
        path = index.data(self.Roles.FilePathRole)
        return f"{protocol}://{path}"

    def get_file_content(self, index_or_path: str | QtCore.QModelIndex):
        if isinstance(index_or_path, str):
            index_or_path = self.index(index_or_path)
        protocol_path = self.get_protocol_path(index_or_path)
        # info = self.fs.info(protocol_path)
        with self.fs.open(protocol_path) as file:
            return file.read()

    def fileIcon(self, index):
        tree_item = index.internalPointer()
        if tree_item is None:
            return None
        return _icon_provider.get_icon(core.FileInfo(tree_item.obj["name"]))

    def fileInfo(self, index):
        tree_item = index.internalPointer()
        return None if tree_item is None else core.FileInfo(tree_item.obj["name"])

    def fileName(self, index) -> str:
        tree_item = index.internalPointer()
        return "" if tree_item is None else pathlib.Path(tree_item.obj["name"]).name

    def filePath(self, index) -> str:
        tree_item = index.internalPointer()
        return "" if tree_item is None else tree_item.obj["name"]

    def isDir(self, index):
        tree_item = index.internalPointer()
        return False if tree_item is None else tree_item.obj["type"] == "directory"

    def mkdir(self, index, name: str):
        tree_item = index.internalPointer()
        path = tree_item.obj["name"]
        new_folder = pathlib.Path(path) / name
        self.fs.mkdir(str(new_folder))

    def size(self, index) -> int:
        tree_item = index.internalPointer()
        return 0 if tree_item is None else tree_item.obj["size"]

    def lastModified(self, index) -> int:
        tree_item = index.internalPointer()
        return (
            ""
            if tree_item is None
            else str(datetime.datetime.fromtimestamp(tree_item.obj["mtime"]))
        )

    def permissions(self, index) -> QtCore.QFileDevice.Permission:
        tree_item = index.internalPointer()
        flag = QtCore.QFileDevice.Permission(0)
        if tree_item is None:
            return flag
        val = oct(int(tree_item.obj["mode"]))[-4:]
        for i in core.filedevice.PERMISSIONS.get_list(int(val, 8)):
            flag |= core.filedevice.PERMISSIONS[i]
        return flag

    def remove(self, index) -> bool:
        tree_item = index.internalPointer()
        path = tree_item.obj["name"]
        self.fs.rm_file(path)
        return self.fs.exists(path)

    def rmdir(self, index) -> bool:
        tree_item = index.internalPointer()
        path = tree_item.obj["name"]
        self.fs.rmdir(path)
        return self.fs.exists(path)

    def setRootPath(self, path: os.PathLike | None) -> core.ModelIndex:
        path = os.fspath(path) if path else self.fs.root_marker
        item = self.fs.info(path)
        self.set_root_item(item)

        self.rootPathChanged.emit(path)

    def rootDirectory(self) -> core.Dir:
        return core.Dir(self.rootPath())

    def rootPath(self) -> str:
        return self._root_item["name"]

    def rename(self, index, value: str):
        tree_item = index.internalPointer()
        path = tree_item.obj["name"]
        old_name = pathlib.Path(path).name
        parent_folder = pathlib.Path(path).parent
        target = str(parent_folder / value)
        self.fs.mv(path, target)
        exists = self.fs.exists(target)
        if exists:
            self.fileRenamed.emit(parent_folder, old_name, value)
        return exists

    def type(self, index) -> str:
        tree_item = index.internalPointer()
        return tree_item.obj["type"]

    def mimeTypes(self):
        return ["text/uri-list"]

    def mimeData(self, indexes):
        mime_data = core.MimeData()
        paths = [i for idx in indexes if (i := idx.data(self.Roles.FilePathRole))]
        urls = [core.Url(i) for i in paths]
        mime_data.setUrls(urls)
        return mime_data

    def supportedDropActions(self):
        return (
            QtCore.Qt.DropAction.MoveAction
            | QtCore.Qt.DropAction.CopyAction
            | QtCore.Qt.DropAction.LinkAction
        )

    def flags(self, index):
        flags = super().flags(index)
        return flags | constants.DROP_ENABLED | constants.DRAG_ENABLED

    # def dragEnterEvent(self, event):
    #     event.accept() if event.mimeData().hasUrls() else super().dragEnterEvent(event)

    def hasChildren(self, parent: core.ModelIndex | None = None):
        parent = core.ModelIndex() if parent is None else parent
        if parent.column() > 0:
            return False
        item = self.data_by_index(parent)
        if self._show_root and item == self._root_item:
            return True
        return item.obj["type"] == "directory"

    def removeRows(self, row: int, count: int, parent):
        print(row, count, parent.data())
        # end_row = row + count - 1
        # with self.remove_rows(row, end_row, parent):
        #     for i in range(end_row, row - 1, -1):
        #         self.items.pop(i)
        return True

    def remove_items(self, offsets: Iterable[int]):
        for offset in sorted(offsets, reverse=True):
            self.removeRow(offset)

    def add_items(
        self,
        items: Iterable,
        position: int | None = None,
        parent: QtCore.QModelIndex | None = None,
    ):
        if position is None:
            position = len(self.items)
        items = list(items)
        # with self.insert_rows(position, position + len(items) - 1, parent):
        #     for i in range(len(items)):
        #         self.items.insert(i + position, items[i])
        #         pass
        #     self.items.extend(items)
        return items

    def canDropMimeData(
        self,
        mime_data: QtCore.QMimeData,
        action: QtCore.Qt.DropAction,
        row: int,
        column: int,
        parent_index: QtCore.QModelIndex,
    ):
        return column == 0 and mime_data.hasFormat("text/uri-list")

    def dropMimeData(
        self,
        mime_data: QtCore.QMimeData,
        action: QtCore.Qt.DropAction,
        row: int,
        column: int,
        parent_index: QtCore.QModelIndex,
    ):
        if not self.canDropMimeData(mime_data, action, row, column, parent_index):
            return False
        print(mime_data.urls(), action, row, column, parent_index.data())
        urls = [core.Url(i) for i in mime_data.urls()]
        if not urls:
            return False
        if action == QtCore.Qt.DropAction.MoveAction:
            with self.change_layout():
                for i in sorted(urls, reverse=True):
                    # self.fs.mv(i, )
                    print("move", i, "to", parent_index.data(self.Roles.FilePathRole))
        elif action == QtCore.Qt.DropAction.CopyAction:
            pass
        elif action == QtCore.Qt.DropAction.LinkAction:
            pass

        column = min(column, 0)
        if row != -1:
            begin_row = row
        elif parent_index.isValid():
            begin_row = self.rowCount(parent_index) + 1
        else:
            begin_row = self.rowCount() + 1
        print(begin_row)
        return True

        # if action == Qt.CopyAction:
        #     # Behavior if parent is a text item
        #     # For example, we select a text and do: CTRL+C CTRL+V
        #     if parent.isValid() and not parent.internalPointer().isFolder():
        #         # We insert copy in parent folder, just below
        #         begin_row = parent.row() + 1
        #         parent = parent.parent()

        #     if parent.isValid() and parent.internalPointer().isFolder():
        #         pass
        # if not items:
        #     return False

        # if action == Qt.CopyAction:
        #     pass

        # r = self.insertItems(items, begin_row, parent)
        # return r

        # return True

    # @functools.singledispatchmethod
    def index(
        self,
        path_or_row: str | int,
        column: int = 0,
        index: QtCore.QModelIndex | None = None,
    ) -> core.ModelIndex:
        if isinstance(path_or_row, int):
            return super().index(path_or_row, column, index)
        return self._iter_path(path_or_row, column, index)

    def _iter_path(
        self, target: os.PathLike, column: int = 0, parent: core.ModelIndex | None = None
    ) -> core.ModelIndex:
        parent = parent or core.ModelIndex()
        target = pathlib.Path(target)
        row_count = self.rowCount(parent)
        for i in range(row_count):
            index = self.index(i, column, parent)
            path = self.filePath(index)
            path = pathlib.Path(path)
            if target == path:
                return index
            elif target.is_relative_to(path):
                if self.canFetchMore(index):
                    self.fetchMore(index)
                return self._iter_path(target, column, index)
        return core.ModelIndex()

    def roleNames(self) -> dict[int, QtCore.QByteArray]:
        return {
            259: QtCore.QByteArray(b"filePermissions"),
            1: QtCore.QByteArray(b"fileIcon"),
            4: QtCore.QByteArray(b"statusTip"),
            3: QtCore.QByteArray(b"toolTip"),
            2: QtCore.QByteArray(b"edit"),
            257: QtCore.QByteArray(b"filePath"),
            0: QtCore.QByteArray(b"display"),
            258: QtCore.QByteArray(b"fileName"),
            5: QtCore.QByteArray(b"whatsThis"),
        }

    def setReadOnly(self, val: bool):
        return NotImplemented

    def isReadOnly(self) -> bool:
        return NotImplemented

    def setOption(self, opt, val):
        return NotImplemented

    def options(self):
        return NotImplemented

    def testOption(self, opt):
        return NotImplemented

    def resolveSymlinks(self, val):
        return NotImplemented

    def setResolveSymlinks(self, val):
        return NotImplemented

    def setNameFilterDisables(self, value):
        return NotImplemented

    def nameFilterDisables(self):
        return NotImplemented

    def setNameFilters(self, filters):
        return NotImplemented

    def nameFilters(self):
        return NotImplemented

    def setFilter(self, filter):
        return NotImplemented

    def filter(self):
        return NotImplemented


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()

    model = FSSpecTreeModel("file", org="phil65", repo="prettyqt", path="/")
    # model.set_root_path("/")
    tree = widgets.TreeView()
    tree.setRootIsDecorated(True)
    tree.setup_dragdrop_move()
    tree.setAlternatingRowColors(True)
    tree.set_model(model)
    tree.set_selection_behavior("rows")
    tree.setUniformRowHeights(True)
    tree.setAnimated(True)
    tree.show()
    with app.debug_mode():
        app.main_loop()



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtQml
from prettyqt.utils import datatypes, get_repr


class QmlPropertyMapMixin(core.ObjectMixin):
    pass


class QmlPropertyMap(QmlPropertyMapMixin, QtQml.QQmlPropertyMap):
    def __repr__(self):
        return get_repr(self, self.as_dict())

    def __setitem__(self, key: str, value: datatypes.Variant):
        self.insert(key, value)

    def __getitem__(self, key: str) -> datatypes.Variant:
        return self.value(key)

    def __delitem__(self, key: str):
        self.clear(key)

    def __contains__(self, key):
        return self.contains(key)

    def __bool__(self):
        return not self.isEmpty()

    def __iter__(self):
        return iter(self.as_dict())

    def items(self):
        return self.as_dict().items()

    def as_dict(self) -> dict[str, datatypes.Variant]:
        return {i: self.value(i) for i in self.keys()}


if __name__ == "__main__":
    propmap = QmlPropertyMap()
    propmap["a"] = 2



from __future__ import annotations

from prettyqt import eventfilters
from prettyqt.qt import QtCore, QtWidgets

CC = QtWidgets.QStyle.ComplexControl
SC = QtWidgets.QStyle.SubControl


class SliderMoveToMouseClickEventFilter(eventfilters.BaseEventFilter):
    def _move_to_mouse_position(self, scrollbar, point: QtCore.QPoint):
        opt = widgets.StyleOptionSlider()
        scrollbar.initStyleOption(opt)
        control = scrollbar.style().hitTestComplexControl(
            CC.CC_ScrollBar, opt, point, scrollbar
        )
        if control not in {SC.SC_ScrollBarAddPage, SC.SC_ScrollBarSubPage}:
            return
        # scroll here
        gr = scrollbar.style().subControlRect(
            CC.CC_ScrollBar, opt, SC.SC_ScrollBarGroove, scrollbar
        )
        sr = scrollbar.style().subControlRect(
            CC.CC_ScrollBar, opt, SC.SC_ScrollBarSlider, scrollbar
        )
        if scrollbar.orientation() == QtCore.Qt.Orientation.Horizontal:
            pos = point.x()
            slider_length = sr.width()
            slider_min = gr.x()
            slider_max = gr.right() - slider_length + 1
            if scrollbar.layoutDirection() == QtCore.Qt.LayoutDirection.RightToLeft:
                opt.upsideDown = not opt.upsideDown
        else:
            pos = point.y()
            slider_length = sr.height()
            slider_min = gr.y()
            slider_max = gr.bottom() - slider_length + 1
        scrollbar.setValue(
            QtWidgets.QStyle.sliderValueFromPosition(
                scrollbar.minimum(),
                scrollbar.maximum(),
                pos - slider_min - slider_length // 2,
                slider_max - slider_min,
                opt.upsideDown,
            )
        )

    def eventFilter(self, source, event):
        match event.type():
            case QtCore.QEvent.Type.MouseMove:
                if event.buttons() & QtCore.Qt.MouseButton.LeftButton:
                    point = event.position().toPoint()
                    self._move_to_mouse_position(source, point)
            case QtCore.QEvent.Type.MouseButtonPress:
                if event.button() == QtCore.Qt.MouseButton.LeftButton:
                    point = event.position().toPoint()
                    self._move_to_mouse_position(source, point)
        return False


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    widget = widgets.PlainTextEdit("gfdgdf\n" * 1000)
    eventfilter = SliderMoveToMouseClickEventFilter(widget.v_scrollbar)
    widget.v_scrollbar.installEventFilter(eventfilter)
    widget.show()
    app.main_loop()



from __future__ import annotations

import logging

from prettyqt import constants, core, eventfilters, widgets
from prettyqt.qt import QtCore


logger = logging.getLogger(__name__)


class DebugMode(eventfilters.BaseEventFilter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.list = widgets.ListWidget()
        self.list.setMouseTracking(True)
        self.list.entered.connect(self._on_entered)
        self.list.itemClicked.connect(self._on_clicked)
        self.frame = widgets.Frame()
        self.frame.set_frame_shape("box")
        self.frame.set_frame_shadow("plain")
        # self.frame.setGeometry(0, 0, 1920, 1080)
        self.frame.setObjectName("testframe")
        self.frame.setStyleSheet("#testframe {border: 5px solid green;}")
        self.frame.setWindowFlags(
            QtCore.Qt.WindowType.FramelessWindowHint
            | QtCore.Qt.WindowType.Tool
            | QtCore.Qt.WindowType.WindowTransparentForInput
            | QtCore.Qt.WindowType.WindowDoesNotAcceptFocus
            | QtCore.Qt.WindowType.WindowStaysOnTopHint
        )
        # self.frame.setWindowOpacity(0.5)
        # self.mask = widgets.Widget()
        # self.mask.set_background_color("green")

    def eventFilter(self, source: QtCore.QObject, event: QtCore.QEvent) -> bool:
        match event.type():
            case core.Event.Type.MouseButtonPress:
                mods = widgets.Application.query_keyboard_modifiers()
                if "ctrl" in mods:
                    pos = source.mapToGlobal(event.pos())
                    candidates = widgets.Application.widgets_at(pos)
                    self.list.clear()
                    for candidate in candidates:
                        self.list.add_item(f"{candidate!r}", data={"user": candidate})
                    self.list.show()
                    return True
                elif "alt" in mods:
                    from prettyqt import ipython

                    console = ipython.InProcessIPythonWidget(self)
                    console.show()
                    pos = source.mapToGlobal(event.pos())
                    widgets.Application.sleep(1)
                    console.push_vars(
                        dict(
                            widgets=widgets.Application.widgets_at(pos), app=widgets.app()
                        )
                    )
                    console.execute("print(widgets)")
        return False

    def _on_clicked(self, item):
        from prettyqt.custom_widgets import widgeteditor

        logger.debug(f"clicked on {item}")
        widget = item.get_data("user")
        self.editor = widgeteditor.WidgetEditor(widget)
        self.list.hide()
        self.editor.show()

    def _on_entered(self, index):
        widget = index.data(constants.USER_ROLE)
        # logger.info(widget)
        for item in self.list:
            candidate = item.get_data("user")
            if candidate != widget:
                pass
                # effect = widgets.GraphicsOpacityEffect(candidate, opacity=0.5)
            else:
                # effect = widgets.GraphicsOpacityEffect(candidate, opacity=0.7)
                # self.mask.setParent(widget)
                # self.mask.setGeometry(widget.geometry())
                # self.mask.show()
                # print("hfd")
                # rect = widget.contentsRect()
                # top_left = rect.topLeft()
                # bottom_right = rect.bottomRight()
                # top_left = widget.mapToGlobal(top_left)
                # bottom_right = widget.mapToGlobal(bottom_right)

                # rect = core.Rect(top_left, bottom_right)
                # whole_frame_region = gui.Region(rect)
                # inner_region = gui.Region(rect.marginsRemoved(core.Margins(5, 5, 5, 5)))
                # inner_frame_region = whole_frame_region.subtracted(inner_region)
                # self.frame.setMask(inner_frame_region)
                self.frame.resize(widget.size())
                self.frame.setParent(widget)
                self.frame.show()

        # with widget.edit_palette() as palette:
        #     palette.setColor(
        #         gui.Palette.ColorRole.Base,
        #         widget.palette().color(gui.Palette.ColorRole.Highlight),
        #     )


if __name__ == "__main__":
    app = widgets.app()
    container = widgets.Widget()
    container.set_layout("horizontal")
    w = widgets.PlainTextEdit(parent=container)
    w2 = widgets.PlainTextEdit(parent=container)
    container.box.add(w)
    container.box.add(w2)
    container.show()
    with app.debug_mode():
        app.sleep(1)
        app.main_loop()
    print(w.get_properties())



from __future__ import annotations

from prettyqt import core
from prettyqt.utils import treeitem


class TreeModel(core.AbstractItemModel):
    def __init__(self, obj, show_root: bool = True, **kwargs):
        super().__init__(**kwargs)
        self._root_item = treeitem.TreeItem(obj=obj)
        self._show_root = show_root
        self.set_root_item(obj)

    def root_index(self) -> core.ModelIndex:  # TODO: needed?
        """Return the index that returns the root element (same as an invalid index)."""
        return core.ModelIndex()

    @property
    def show_root(self):
        """Return True if the inspected node is visible.

        In that case an invisible root node has been added.
        """
        return self._show_root

    def set_root_item(self, obj):
        if self._show_root:
            self._root_item = treeitem.TreeItem(obj=None)
            self._root_item.children_fetched = True
            self.inspected_item = treeitem.TreeItem(obj=obj)
            self._root_item.append_child(self.inspected_item)
            # root_index = self.index(0, 0)
            # self.fetchMore(self.index(0, 0, root_index))
        else:
            # The root itself will be invisible
            self._root_item = treeitem.TreeItem(obj=obj)
            self.inspected_item = self._root_item

            # Fetch all items of the root so we can select the first row in the ctor.
            root_index = self.index(0, 0)
            self.fetchMore(root_index)

    @property
    def root_item(self) -> treeitem.TreeItem:
        """Return the root ObjectBrowserTreeItem."""
        return self._root_item

    def data_by_index(self, index: core.ModelIndex) -> treeitem.TreeItem:
        return index.internalPointer() if index.isValid() else self.root_item

    def index(
        self, row: int, column: int, parent: core.ModelIndex | None = None
    ) -> core.ModelIndex:
        parent = parent or core.ModelIndex()
        parent_item = self.data_by_index(parent)

        if not self.hasIndex(row, column, parent):
            return core.ModelIndex()

        if child_item := parent_item.child(row):  # isnt this always true?
            return self.createIndex(row, column, child_item)
        return core.ModelIndex()

    def parent(self, index: core.ModelIndex) -> core.ModelIndex:  # type:ignore
        if not index.isValid():
            return core.ModelIndex()

        child_item = self.data_by_index(index)
        parent_item = child_item.parent()  # type: ignore

        if parent_item is None or parent_item == self.root_item:
            return core.ModelIndex()

        return self.createIndex(parent_item.row(), 0, parent_item)

    def rowCount(self, parent: core.ModelIndex | None = None):
        parent = core.ModelIndex() if parent is None else parent
        return 0 if parent.column() > 0 else self.data_by_index(parent).child_count()

    def hasChildren(self, parent: core.ModelIndex | None = None):
        parent = core.ModelIndex() if parent is None else parent
        return 0 if parent.column() > 0 else self.data_by_index(parent).has_children

    def canFetchMore(self, parent: core.ModelIndex | None = None):
        parent = core.ModelIndex() if parent is None else parent
        if parent.column() > 0:
            return 0
        else:
            return not self.data_by_index(parent).children_fetched

    def fetchMore(self, parent: core.ModelIndex | None = None):
        """Fetch the children given the model index of a parent node.

        Adds the children to the parent.
        """
        parent = core.ModelIndex() if parent is None else parent
        if parent.column() > 0:
            return

        parent_item = self.data_by_index(parent)
        if parent_item.children_fetched:
            return

        tree_items = self._fetch_object_children(parent_item)

        with self.insert_rows(0, len(tree_items) - 1, parent):
            for tree_item in tree_items:
                parent_item.append_child(tree_item)
            parent_item.children_fetched = True

    def _fetch_object_children(self, treeitem) -> list:
        return NotImplemented



from __future__ import annotations

from collections.abc import Mapping

from prettyqt import core, custom_models
from prettyqt.utils import treeitem


class MappingModel(custom_models.ColumnItemModel):
    """Model that provides an interface to an objectree that is build of tree items."""

    def __init__(
        self,
        obj: list[dict],
        nested_key=None,
        **kwargs,
    ):
        super().__init__(
            obj=obj,
            columns=[],
            show_root=False,
            **kwargs,
        )
        self.nested_key = nested_key
        columns = []
        for k, v in obj[0].items():

            def get_repr(x, k=k, v=v):
                return repr(v) if isinstance(x.obj, Mapping) else repr(x.obj)

            col = custom_models.ColumnItem(
                name=k,
                doc=k,
                label=get_repr,
            )
            columns.append(col)
        self.set_columns(columns)

    def hasChildren(self, parent: core.ModelIndex | None = None):
        parent = core.ModelIndex() if parent is None else parent
        if parent.column() > 0:
            return False
        item = self.data_by_index(parent)
        if item in [self._root_item, self.inspected_item]:
            return True
        if self.nested_key is None:
            return False
        return isinstance(item.obj.get(self.nested_key), Mapping)

    def _fetch_object_children(self, item: treeitem.TreeItem) -> list[treeitem.TreeItem]:
        match item.obj:
            # case Mapping():
            #     return [
            #         treeitem.TreeItem(obj=v)
            #         for k, v in item.obj.items()
            #     ]
            case list():
                return [treeitem.TreeItem(obj=item) for item in item.obj]


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    dct = dict(a=2, b="hallo", c="hall")
    dct2 = dict(a=2, b="hallo2", c="hallo3")
    model = MappingModel([dct, dct2])
    table = widgets.TreeView()
    table.setRootIsDecorated(True)
    # table.setSortingEnabled(True)
    table.set_model(model)
    table.show()
    app.main_loop()



from __future__ import annotations

from prettyqt import constants, custom_models, gui


COL_NAME = custom_models.ColumnItem(
    name="Name",
    doc="WhatsThis",
    label=lambda x: x.whatsThis(),
    # decoration=lambda x: x.icon(),
)

COL_ENABLED = custom_models.ColumnItem(
    name="Enabled",
    doc="Enabled",
    checkable=True,
    checkstate=lambda x: x.isEnabled(),
    set_checkstate=lambda item, value: item.setEnabled(not item.isEnabled()),
)

COL_SHORTCUT = custom_models.ColumnItem(
    name="Shortcut",
    doc="Shortcut",
    label=lambda x: x.key().toString(),
)

COL_CONTEXT = custom_models.ColumnItem(
    name="Context",
    doc="Context",
    label=lambda x: constants.SHORTCUT_CONTEXT.inverse[x.context()],
)

COL_AUTOREPEAT = custom_models.ColumnItem(
    name="AutoRepeat",
    doc="AutoRepeat",
    checkable=True,
    checkstate=lambda item: item.autoRepeat(),
    set_checkstate=lambda item, value: item.setAutoRepeat(not item.autoRepeat()),
)
COL_PARENT = custom_models.ColumnItem(
    name="Parent widget",
    doc="Parent widget",
    label=lambda item: item.parent().windowTitle() or item.parent().__class__.__name__,
)


COLUMNS = [
    COL_NAME,
    COL_ENABLED,
    COL_SHORTCUT,
    COL_CONTEXT,
    COL_AUTOREPEAT,
    COL_PARENT,
]


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    view = widgets.TreeView()
    shortcuts = [
        gui.Shortcut(
            "Ctrl+A",
            view,
            lambda: print("test"),
            object_name="super duper shortcut",
        ),
        gui.Shortcut(
            "Ctrl+B",
            view,
            print,
            object_name="this is an shortcut",
        ),
        # gui.Shortcut(
        #     view,
        #     object_name="another one",
        #     key="Ctrl+Alt+A",
        # ),
    ]
    for i in shortcuts:
        i.setWhatsThis("abc")
    model = custom_models.ColumnTableModel(shortcuts, COLUMNS, parent=view)
    view.setModel(model)
    view.resize(640, 480)
    view.setEditTriggers(view.EditTrigger.AllEditTriggers)
    view.set_selection_behavior("rows")
    view.adapt_sizes()
    view.show()
    with app.debug_mode():
        app.main_loop()



from __future__ import annotations

from typing import Literal

from prettyqt import bluetooth, core
from prettyqt.qt import QtBluetooth, QtCore
from prettyqt.utils import bidict


Error = QtBluetooth.QBluetoothServer.Error

ERROR = bidict(
    none=Error.NoError,
    unknown=Error.UnknownError,
    powered_off=Error.PoweredOffError,
    input_output=Error.InputOutputError,
    service_already_registered=Error.ServiceAlreadyRegisteredError,
    unsupported_protocol=Error.UnsupportedProtocolError,
    missing_permissions=Error.MissingPermissionsError,
)

ErrorStr = Literal[
    "none",
    "unknown",
    "powered_off",
    "input_output",
    "service_already_registered",
    "unsupported_protocol",
    "missing_permissions",
]


class BluetoothServer(core.ObjectMixin, QtBluetooth.QBluetoothServer):
    def __init__(
        self,
        protocol: bluetooth.bluetoothserviceinfo.ProtocolStr
        | QtBluetooth.QBluetoothServiceInfo.Protocol,
        parent: QtCore.QObject | None = None,
    ):
        if isinstance(protocol, str):
            protocol = bluetooth.bluetoothserviceinfo.PROTOCOL[protocol]
        super().__init__(protocol, parent)

    def get_error(self) -> ErrorStr:
        return ERROR.inverse[self.error()]

    def get_server_type(self) -> bluetooth.bluetoothserviceinfo.ProtocolStr:
        return bluetooth.bluetoothserviceinfo.PROTOCOL.inverse[self.serverType()]

    def get_server_address(self) -> bluetooth.BluetoothAddress:
        return bluetooth.BluetoothAddress(self.serverAddress())


if __name__ == "__main__":
    from prettyqt import core

    app = core.app()
    server = BluetoothServer("l2_cap")
    with app.debug_mode():
        app.main_loop()



from __future__ import annotations

import logging
from typing import Literal

from prettyqt import constants, core, gui
from prettyqt.qt import QtCore, QtGui
from prettyqt.utils import InvalidParamError, serializemixin

logger = logging.getLogger(__name__)


class Cursor(serializemixin.SerializeMixin, QtGui.QCursor):
    @classmethod
    def fake_mouse_move(cls):
        cls.setPos(cls.pos() + core.Point(0, 1))
        gui.Application.processEvents()
        cls.setPos(cls.pos() - core.Point(0, 1))

    @classmethod
    def click(cls, key=QtCore.Qt.MouseButton.LeftButton):
        from prettyqt import widgets

        app = widgets.app()
        widget = app.widgetAt(cls.pos())
        if widget is None:
            return
        pos = cls.pos().toPointF()
        local = pos - widget.mapToGlobal(core.PointF(0, 0))
        logger.info(f"sending MouseClick events to {widget} at {local}")
        event = gui.MouseEvent(
            QtCore.QEvent.Type.MouseButtonPress,
            local,
            pos,
            key,
            QtCore.Qt.MouseButton.NoButton,
            QtCore.Qt.KeyboardModifier(0),
        )

        core.CoreApplication.sendEvent(widget, event)
        event = gui.MouseEvent(
            QtCore.QEvent.Type.MouseButtonRelease,
            local,
            pos,
            key,
            QtCore.Qt.MouseButton.NoButton,
            QtCore.Qt.KeyboardModifier(0),
        )

        core.CoreApplication.sendEvent(widget, event)

    def set_shape(self, shape: constants.CursorShapeStr):
        """Set cursor shape.

        Args:
            shape: shape to use

        Raises:
            InvalidParamError: shape does not exist
        """
        if shape not in constants.CURSOR_SHAPE:
            raise InvalidParamError(shape, constants.CURSOR_SHAPE)
        self.setShape(constants.CURSOR_SHAPE[shape])

    def get_shape(self) -> constants.CursorShapeStr:
        """Return current cursor shape.

        Returns:
            cursor shape
        """
        return constants.CURSOR_SHAPE.inverse[self.shape()]

    @classmethod
    def get_pos(cls) -> core.Point:
        return core.Point(cls.pos())

    @classmethod
    def set_pos(
        cls,
        where: Literal["screen", "current"] | QtGui.QScreen
        # | QtWidgets.QWidget
        | QtCore.QRect | QtCore.QPoint | tuple[int, int] | tuple[int, int, int, int],
        how: Literal[
            "center",
            "top",
            "left",
            "bottom",
            "right",
            "top_left",
            "top_right",
            "bottom_left",
            "bottom_right",
        ] = "center",
        x_offset: int = 0,
        y_offset: int = 0,
        duration: int = 0,
    ):
        """Position cursor onto screen position / widget / window / screen.

        Arguments:
            where: where to positin on
            how: How to align
            x_offset: additional x offset for final position
            y_offset: additional y offset for final position
            duration: movement time
        """
        match where:
            case "current":
                p = cls.pos()
                geom = core.Rect(p, p)
            case QtCore.QPoint():
                geom = core.Rect(where, where)
            case (int(), int()):
                p = core.Point(*where)
                geom = core.Rect(p, p)
            case (int(), int(), int(), int()):
                geom = core.Rect(*where)
            case QtCore.QRect():
                geom = where
            case "screen":
                geom = gui.GuiApplication.primaryScreen().geometry()
            case QtGui.QScreen():
                geom = where.geometry()
            case _:  # not wanting to import QtWidgets here... perhaps create a protocol.
                geom = where.frameGeometry()
        match how:
            case "center":
                new = geom.center()
            case "top":
                new = core.Point(geom.center().x(), geom.top())
            case "bottom":
                new = core.Point(geom.center().x(), geom.bottom())
            case "left":
                new = core.Point(geom.left(), geom.center().y())
            case "right":
                new = core.Point(geom.right(), geom.center().y())
            case "top_right":
                new = geom.topRight()
            case "top_left":
                new = geom.topLeft()
            case "bottom_right":
                new = geom.bottomRight()
            case "bottom_left":
                new = geom.bottomLeft()
        new_pos = core.Point(new.x() + x_offset, new.y() + y_offset)
        if duration > 0:
            from prettyqt.custom_animations import cursormoveanimation

            cls._cursor_animation = cursormoveanimation.CursorMoveAnimation(
                duration=duration, end_value=new_pos
            )
            cls._cursor_animation.start()
        else:
            cls.setPos(new_pos)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    widget = widgets.PushButton()
    widget.show()
    with app.debug_mode():
        app.sleep(2)
        widget.clicked.connect(lambda: logger.info("x"))
        Cursor.set_pos("screen", duration=1000)
        app.sleep(2)
        Cursor.set_pos(widget, how="right")



from __future__ import annotations

from prettyqt import bluetooth
from prettyqt.qt import QtBluetooth


class BluetoothHostInfo(QtBluetooth.QBluetoothHostInfo):
    def set_address(self, address: QtBluetooth.QBluetoothAddress | int | str):
        if isinstance(address, int | str):
            address = QtBluetooth.QBluetoothAddress(address)
        self.setAddress(address)

    def get_address(self) -> bluetooth.BluetoothAddress:
        return bluetooth.BluetoothAddress(self.address())


if __name__ == "__main__":
    address = BluetoothHostInfo()
    print(address)



from __future__ import annotations

from collections.abc import Callable
import functools
import logging

from prettyqt import core
from prettyqt.qt import QtCore


logger = logging.getLogger(__name__)


class Signals(core.Object):
    finished = core.Signal()
    error = core.Signal(Exception)
    result = core.Signal(object)
    progress = core.Signal(int)


class Worker(core.Runnable):
    """Worker thread.

    Inherits from QRunnable to handler worker thread setup, signals and wrap-up.

    Arguments:
        fn:     The function callback to run on this worker thread.
                Supplied args and kwargs will be passed through to the runner.
        args:   Arguments to pass to the callback function
        kwargs: Keywords to pass to the callback function

    """

    def __init__(self, fn: Callable, *args, **kwargs):
        self.fn = fn
        self.args = args
        self.keyword_args = kwargs
        self.signals = Signals()
        super().__init__()

        # Add the callback to our kwargs
        # self.keyword_args["progress_callback"] = self.signals.progress

    @core.Slot()
    def run(self):
        """Initialise the runner function with passed args, kwargs."""
        try:
            result = self.fn(*self.args, **self.keyword_args)
            self.signals.result.emit(result)
            self.signals.finished.emit()
        except Exception as e:
            logger.exception(e)
            self.signals.error.emit(e)


class GeneratorWorker(Worker):
    @core.Slot()
    def run(self):
        try:
            for result in self.fn(*self.args, **self.keyword_args):
                self.signals.result.emit(result)
            self.signals.finished.emit()
        except Exception as e:
            logger.exception(e)
            self.signals.error.emit(e)


def run_async(func: Callable) -> Callable:
    @functools.wraps(func)
    def async_func(*args, **kwargs):
        worker = Worker(func, *args, **kwargs)
        pool = ThreadPool.globalInstance()
        pool.start(worker)

    return async_func


class ThreadPool(core.ObjectMixin, QtCore.QThreadPool):
    def __contains__(self, other: QtCore.QThread):
        return self.contains(other)

    def get_thread_priority(self) -> core.thread.PriorityStr:
        return core.thread.PRIORITY.inverse[self.priority()]

    def set_thread_priority(self, priority: core.thread.PriorityStr):
        prio = core.thread.PRIORITY[self.threadPriority()]
        self.setThreadPriority(prio)


if __name__ == "__main__":
    pool = ThreadPool()



"""quick module.

contains QtQuick-based classes
"""

from .quicktextdocument import QuickTextDocument
from .quickrendercontrol import QuickRenderControl
from .quickimageresponse import QuickImageResponse
from .quickimageprovider import QuickImageProvider, QuickImageProviderMixin
from .quickasyncimageprovider import QuickAsyncImageProvider
from .quickitemgrabresult import QuickItemGrabResult
from .quickitem import QuickItem, QuickItemMixin
from .quickpainteditem import QuickPaintedItem
from .quickwindow import QuickWindow
from .quickview import QuickView
from .sgnode import SGNode


__all__ = [
    "SGNode",
    "QuickView",
    "QuickItem",
    "QuickItemMixin",
    "QuickWindow",
    "QuickPaintedItem",
    "QuickTextDocument",
    "QuickRenderControl",
    "QuickImageResponse",
    "QuickImageProvider",
    "QuickImageProviderMixin",
    "QuickAsyncImageProvider",
    "QuickItemGrabResult",
]



from __future__ import annotations

from collections import defaultdict
from collections.abc import Callable, Container
import contextlib

# import inspect
import itertools
import logging
from typing import TYPE_CHECKING, Any, TypeVar

from typing_extensions import Self

from prettyqt import constants, core
from prettyqt.qt import QtCore
from prettyqt.utils import datatypes, get_repr, helpers


if TYPE_CHECKING:
    from prettyqt import eventfilters

T = TypeVar("T", bound=QtCore.QObject)

counter_dict: defaultdict = defaultdict(itertools.count)

logger = logging.getLogger(__name__)


class ObjectMixin:
    Properties: dict[type, list[str]] = {}

    def __init__(self, *args, **kwargs):
        self._eventfilters = set()
        new = {}
        if kwargs:
            mapper = self._get_map()
            for k, v in kwargs.items():
                # this allows snake_case naming.
                camel_k = helpers.to_lower_camel(k)
                # allow str values instead of enum
                if camel_k in mapper and isinstance(v, str):
                    new[camel_k] = mapper[camel_k][v]
                # allow str values for common icon kwargs
                elif k in {"window_icon", "icon"} and isinstance(v, str):
                    from prettyqt import iconprovider

                    new[camel_k] = iconprovider.get_icon(v)
                # regular case
                else:
                    new[camel_k] = v
        super().__init__(*args, **new)

    def _get_map(self):
        """Can be implemented by subclasses to support str -> Enum conversion.

        To get data from all subclasses, we always fetch _get_map from super(),
        append our own shit and return it.
        """
        return {}

    def __repr__(self):  # we already monkeypatch QObject
        return get_repr(self, self.objectName())

    def __setstate__(self, state):
        self.set_properties(state)

    def __getstate__(self):
        return self.get_properties()

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def __getattr__(self, val):
        cameled = helpers.to_lower_camel(val)
        if cameled in dir(self):
            return getattr(self, cameled)
        raise AttributeError(val)

    def installEventFilter(self, eventfilter: QtCore.QObject):
        if eventfilter in self._eventfilters:
            logger.warning("Trying to install same EventFilter multiple times.")
            return
        self._eventfilters.add(eventfilter)
        super().installEventFilter(eventfilter)

    def removeEventFilter(self, eventfilter: QtCore.QObject):
        if eventfilter not in self._eventfilters:
            logger.warning("Trying to remove non-installed EventFilter.")
            return
        self._eventfilters.remove(eventfilter)
        super().removeEventFilter(eventfilter)

    def add_callback_for_event(
        self,
        callback: Callable[[QtCore.QEvent], bool],
        include: QtCore.QEvent.Type | Container[QtCore.QEvent.Type] | None = None,
        exclude: QtCore.QEvent.Type | Container[QtCore.QEvent.Type] | None = None,
    ) -> eventfilters.EventCatcher:
        """Connect widget events to a callback.

        if include is set, it behaves like a whitelist.
        if exclude is set, it behaves like a blacklist.
        The QEvent is passed to the callback as an argument, and the callback
        needs to return True or False to indicate whether the Event should be filtered.

        Arguments:
            callback: Callback to execute when event is triggered
            include: Events to include
            exclude: Events to exclude
        """
        from prettyqt import eventfilters

        eventfilter = eventfilters.EventCatcher(include, exclude, callback, parent=self)
        self.installEventFilter(eventfilter)
        return eventfilter

    def serialize_fields(self):
        return dict(object_name=self.objectName())

    def serialize(self) -> dict[str, Any]:
        return self.get_properties()
        # dct = {}
        # for klass in reversed(inspect.getmro(type(self))):
        #     if "serialize_fields" in klass.__dict__:
        #         data = klass.serialize_fields(self)  # type: ignore
        #         dct |= data
        # return dct

    @contextlib.contextmanager
    def block_signals(self):
        blocked = self.blockSignals(True)
        yield None
        self.blockSignals(blocked)

    def to_json(self):
        dct = self.__getstate__()
        for k, v in dct.items():
            if isinstance(v, ObjectMixin):
                dct[k] = v.to_json()
        return dct

    def set_unique_id(self):
        class_name = type(self).__name__
        count = next(counter_dict[class_name])
        self.set_id(f"{class_name}_{count}")

    def set_id(self, name: str):
        self.setObjectName(name)

    def get_id(self) -> str:
        return self.objectName()

    def has_id(self) -> bool:
        return self.objectName() != ""

    @classmethod
    def get_static_metaobject(cls) -> core.MetaObject:
        return core.MetaObject(cls.staticMetaObject)

    def get_metaobject(self) -> core.MetaObject:
        return core.MetaObject(self.metaObject())

    # @property
    # def id(self) -> str:
    #     return self.objectName()

    # @id.setter
    # def id(self, name: str):
    #     self.setObjectName(name)

    def find_children(
        self,
        typ: type[T] = QtCore.QObject,
        name: str | QtCore.QRegularExpression | None = None,
        recursive: bool = True,
    ) -> list[T]:
        if recursive:
            flag = QtCore.Qt.FindChildOption.FindChildrenRecursively
        else:
            flag = QtCore.Qt.FindChildOption.FindDirectChildrenOnly
        return self.findChildren(typ, name=name, options=flag)  # type: ignore

    def find_child(
        self,
        typ: type[T] = QtCore.QObject,
        name: str | QtCore.QRegularExpression | None = None,
        recursive: bool = True,
    ) -> T | None:
        if recursive:
            flag = QtCore.Qt.FindChildOption.FindChildrenRecursively
        else:
            flag = QtCore.Qt.FindChildOption.FindDirectChildrenOnly
        return self.findChild(typ, name, flag)  # type: ignore

    def find_parent(
        self, typ: type[T] = QtCore.QObject, name: str | None = None
    ) -> T | None:
        node = self
        while node:
            node = node.parent()
            if isinstance(node, typ) and (name is None or node.objectName() == name):
                return node
        return None

    def start_timer(
        self,
        interval: int | str,
        timer_type: constants.TimerTypeStr = "coarse",
    ) -> int | None:
        if isinstance(interval, str):
            interval = helpers.parse_time(interval)
        result = self.startTimer(interval, constants.TIMER_TYPE[timer_type])
        return None if result == 0 else result

    def start_callback_timer(
        self,
        callback: Callable,
        interval: int | str,
        single_shot: bool = False,
        timer_type: constants.TimerTypeStr = "coarse",
    ) -> int | None:
        if isinstance(interval, str):
            interval = helpers.parse_time(interval)
        timer = core.Timer(
            self, single_shot=single_shot, interval=interval, timer_type=timer_type
        )
        timer.timeout.connect(callback)
        timer.start()
        return timer

    def get_properties(
        self, include_super: bool = True, cast: bool = True
    ) -> dict[str, Any]:
        metaobj = self.get_metaobject()
        props = metaobj.get_properties(include_super=include_super)
        return {
            i.name(): datatypes.make_serializable(i.read(self))
            for i in props
            if i.get_name() not in ["children", "frameShadow", "state"]
        }

    def set_properties(self, props: dict[str, Any], include_super: bool = True):
        metaobj = self.get_metaobject()
        metaprops = metaobj.get_properties(include_super=include_super)
        for metaprop in metaprops:
            if (name := metaprop.name()) in props:
                value = props[name]
                metaprop.write(self, value)

    def get_dynamic_properties(self) -> dict[str, Any]:
        return {
            i.data().decode(): self.property(i.data().decode())
            for i in self.dynamicPropertyNames()
        }

    def copy(self) -> Self:
        metaobj = self.get_metaobject()
        new = type(self)()
        for prop in metaobj.get_properties():
            if prop.isWritable():
                val = prop.read(self)
                prop.write(new, val)
        # MetaObject can return non-existing signals, dont know why.
        # also filter out duplicates.
        signal_names = {
            s.get_name() for s in metaobj.get_signals() if hasattr(self, s.get_name())
        }
        for signal_name in signal_names:
            own_signal = self.__getattribute__(signal_name)
            new_signal = new.__getattribute__(signal_name)
            # own_signal.connect(new_signal)
            new_signal.connect(own_signal)
            logger.debug(f"connected {signal_name}")
        return new


class Object(ObjectMixin, QtCore.QObject):
    pass


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    obj = core.Object(object_name="jkjk")
    obj.copy()
    meta = obj.get_metaobject()
    prop = meta.get_property(0)
    with app.debug_mode():
        app.sleep(1)



from __future__ import annotations

from prettyqt import custom_models, gui


COL_NAME = custom_models.ColumnItem(
    name="Name",
    doc="Action name",
    label=lambda x: x.text(),
    # decoration=lambda x: x.icon(),
)

COL_DESCRIPTION = custom_models.ColumnItem(
    name="Description",
    doc="Description",
    label=lambda x: x.toolTip(),
)

COL_SHORTCUT = custom_models.ColumnItem(
    name="Shortcut",
    doc="Shortcut",
    label=lambda x: x.shortcut().toString(),
)

COL_PRIORITY = custom_models.ColumnItem(
    name="Priority",
    doc="Priority",
    label=lambda x: gui.action.PRIORITIES.inverse[x.priority()],
)

COL_CHECKSTATE = custom_models.ColumnItem(
    name="Checkstate",
    doc="Checkstate",
    checkstate=lambda x: x.isChecked(),
)

COL_USAGE_COUNT = custom_models.ColumnItem(
    name="Usage count",
    doc="Usage count",
    label=lambda x: x.usage_count if hasattr(x, "usage_count") else 0,
)


COLUMNS = [
    COL_NAME,
    COL_DESCRIPTION,
    COL_SHORTCUT,
    COL_PRIORITY,
    COL_CHECKSTATE,
    COL_USAGE_COUNT,
]


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    view = widgets.TreeView()
    actions = [
        gui.Action(
            text="super duper action",
            shortcut="Ctrl+A",
            tool_tip="some Tooltip text",
            icon="mdi.folder",
            triggered=lambda: print("test"),
        ),
        gui.Action(
            text="this is an action",
            shortcut="Ctrl+B",
            tool_tip="Tooltip",
            icon="mdi.folder-outline",
            checked=True,
            checkable=True,
        ),
        gui.Action(
            text="another one",
            shortcut="Ctrl+Alt+A",
            tool_tip="Some longer tool_tippp",
            icon="mdi.folder",
        ),
        gui.Action(text="a", shortcut="Ctrl+A", tool_tip="Tooltip", icon="mdi.folder"),
    ]
    model = custom_models.ColumnTableModel(actions, COLUMNS, parent=view)
    view.setModel(model)
    view.resize(640, 480)
    view.set_selection_behavior("rows")
    view.adapt_sizes()
    view.show()
    app.main_loop()



from __future__ import annotations

from typing import Literal

from prettyqt import gui
from prettyqt.qt import QtPrintSupport
from prettyqt.utils import InvalidParamError, bidict


COLOR_MODE = bidict(
    color=QtPrintSupport.QPrinter.ColorMode.Color,
    gray_scale=QtPrintSupport.QPrinter.ColorMode.GrayScale,
)

ColorModeStr = Literal["color", "gray_scale"]

DUPLEX_MODE = bidict(
    none=QtPrintSupport.QPrinter.DuplexMode.DuplexNone,
    auto=QtPrintSupport.QPrinter.DuplexMode.DuplexAuto,
    long_side=QtPrintSupport.QPrinter.DuplexMode.DuplexLongSide,
    short_side=QtPrintSupport.QPrinter.DuplexMode.DuplexShortSide,
)

DuplexModeStr = Literal["none", "auto", "long_side", "short_side"]


class Printer(gui.PagedPaintDeviceMixin, QtPrintSupport.QPrinter):
    # def get_source(self) -> pathlib.Path:
    #     return pathlib.Path(self.source().toLocalFile())

    def get_duplex(self) -> DuplexModeStr:
        return DUPLEX_MODE.inverse[self.duplex()]

    def get_pdf_version(self) -> gui.pagedpaintdevice.PdfVersionStr:
        return gui.pagedpaintdevice.PDF_VERSION.inverse[self.pdfVersion()]

    def set_pdf_version(self, version: gui.pagedpaintdevice.PdfVersionStr):
        """Set pdf version.

        Args:
            version: pdf version

        Raises:
            InvalidParamError: pdf version does not exist
        """
        if version not in gui.pagedpaintdevice.PDF_VERSION:
            raise InvalidParamError(version, gui.pagedpaintdevice.PDF_VERSION)
        self.setPdfVersion(gui.pagedpaintdevice.PDF_VERSION[version])



from __future__ import annotations

from collections.abc import Iterator
import contextlib
import sys

from prettyqt import constants, core, gui, iconprovider
from prettyqt.qt import QtCore, QtGui
from prettyqt.utils import InvalidParamError, colors, datatypes


class GuiApplicationMixin(core.CoreApplicationMixin):
    palette_changed = core.Signal(gui.Palette)

    def serialize_fields(self):
        return dict(icon=self.get_icon())

    def event(self, e):
        match e.type():
            case QtCore.QEvent.Type.ApplicationPaletteChange:
                self.palette_changed.emit(gui.Palette(self.palette()))
        return super().event(e)

    @classmethod
    @contextlib.contextmanager
    def override_cursor(cls, cursor: constants.CursorShapeStr):
        cls.set_override_cursor(cursor)
        yield cursor
        cls.restore_override_cursor()

    @classmethod
    def set_override_cursor(cls, cursor: constants.CursorShapeStr):
        crs = gui.Cursor(constants.CURSOR_SHAPE[cursor])
        cls.setOverrideCursor(crs)

    @classmethod
    def restore_override_cursor(cls):
        cls.restoreOverrideCursor()

    @classmethod
    def get_clipboard(cls) -> gui.Clipboard:
        return gui.Clipboard(cls.clipboard())

    @classmethod
    @contextlib.contextmanager
    def edit_palette(cls) -> Iterator[gui.Palette]:
        palette = gui.Palette(cls.palette())
        yield palette
        cls.setPalette(palette)

    def set_layout_direction(self, direction: constants.LayoutDirectionStr):
        """Set layout direction.

        Args:
            direction: layout direction

        Raises:
            InvalidParamError: layout direction does not exist
        """
        if direction not in constants.LAYOUT_DIRECTION:
            raise InvalidParamError(direction, constants.LAYOUT_DIRECTION)
        self.setLayoutDirection(constants.LAYOUT_DIRECTION[direction])

    def get_layout_direction(self) -> constants.LayoutDirectionStr:
        """Get the current layout direction.

        Returns:
            layout direction
        """
        return constants.LAYOUT_DIRECTION.inverse[self.layoutDirection()]

    @classmethod
    def set_high_dpi_scale_factor_rounding_policy(
        cls, policy: constants.HighDpiScaleFactorRoundingPolicyStr
    ):
        """Set high dpi scale factor rounding policy.

        Args:
            policy: rounding policy

        Raises:
            InvalidParamError: rounding policy does not exist
        """
        if policy not in constants.HIGH_DPI_SCALE_FACTOR_ROUNDING_POLICY:
            raise InvalidParamError(
                policy, constants.HIGH_DPI_SCALE_FACTOR_ROUNDING_POLICY
            )
        cls.setHighDpiScaleFactorRoundingPolicy(
            constants.HIGH_DPI_SCALE_FACTOR_ROUNDING_POLICY[policy]
        )

    @classmethod
    def get_high_dpi_scale_factor_rounding_policy(
        cls,
    ) -> constants.HighDpiScaleFactorRoundingPolicyStr:
        """Get the current high dpi scale factor rounding policy.

        Returns:
            rounding policy
        """
        return constants.HIGH_DPI_SCALE_FACTOR_ROUNDING_POLICY.inverse[
            cls.highDpiScaleFactorRoundingPolicy()
        ]

    @classmethod
    def get_application_state(cls) -> list[constants.ApplicationStateStr]:
        """Get the current application state.

        Returns:
            application state
        """
        return [
            k
            for k, v in constants.APPLICATION_STATES.items()
            if v & cls.applicationState()  # type: ignore
        ]

    def get_primary_screen(self) -> gui.Screen:
        return gui.Screen(self.primaryScreen())

    def get_screen_at(self, point: datatypes.PointType) -> gui.Screen:
        if isinstance(point, tuple):
            point = QtCore.QPoint(*point)
        return gui.Screen(self.screenAt(point))

    def get_screens(self) -> list[gui.Screen]:
        return [gui.Screen(i) for i in self.screens()]

    @classmethod
    def get_input_method(cls) -> gui.InputMethod:
        return gui.InputMethod(cls.inputMethod())

    @classmethod
    def copy_to_clipboard(cls, text: str):
        """Sets clipboard to supplied text."""
        cb = cls.clipboard()
        cb.clear(mode=cb.Mode.Clipboard)
        cb.setText(text, mode=cb.Mode.Clipboard)

    @classmethod
    def get_font(cls) -> gui.Font:
        return gui.Font(cls.font())

    def set_icon(self, icon: datatypes.IconType):
        """Set the default window icon.

        Args:
            icon: icon to use
        """
        icon = iconprovider.get_icon(icon, color=colors.WINDOW_ICON_COLOR)
        self.setWindowIcon(icon)

    def get_icon(self) -> gui.Icon | None:
        icon = self.windowIcon()
        return None if icon.isNull() else gui.Icon(self.windowIcon())

    @classmethod
    def set_palette(cls, palette: constants.ThemeStr | QtGui.QPalette):
        if palette == "default":
            pal = gui.Palette()
        elif palette == "dark":
            pal = gui.Palette.create_dark_palette()
        else:
            pal = palette
        cls.setPalette(pal)

    @classmethod
    def get_keyboard_modifiers(cls) -> list[constants.KeyboardModifierStr]:
        return constants.KEYBOARD_MODIFIERS.get_list(cls.keyboardModifiers())

    @classmethod
    def query_keyboard_modifiers(cls) -> list[constants.KeyboardModifierStr]:
        return constants.KEYBOARD_MODIFIERS.get_list(cls.queryKeyboardModifiers())

    @classmethod
    def get_palette(cls) -> gui.Palette:
        return gui.Palette(cls.palette())

    def set_badge_number(self, number: int | None):
        self.setBadgeNumber(number or 0)

    def set_progress_value(self, value: int, total: int = 100):
        windows = self.topLevelWindows()
        if not windows:
            return None
        if sys.platform.startswith("win"):
            from prettyqt.utils.platforms.windows import taskbaritem

            window_id = windows[0].winId()
            tb = taskbaritem.TaskBarItem(window_id)
            tb.set_progress_value(value, total)


class GuiApplication(GuiApplicationMixin, QtGui.QGuiApplication):
    pass


if __name__ == "__main__":
    app = gui.app()
    app.set_badge_number(5)
    app.main_loop()



from __future__ import annotations

import enum

from prettyqt import core, gui, widgets


class EnumFlagWidget(widgets.ToolButton):
    value_changed = core.Signal(enum.Flag)

    def __init__(self, *args, **kwargs):
        self._enum_class = None
        self._action_map = {}
        self.button_menu = widgets.Menu()
        self.button_menu.triggered.connect(self._on_menu_click)
        super().__init__(*args, **kwargs)
        self.setMenu(self.button_menu)
        self.set_popup_mode("instant")

    def _on_menu_click(self):
        self.set_text(" | ".join(i.name for i in self.get_value()))
        self.value_changed.emit(self.get_value())

    def clear(self):
        self._action_map = {}
        self.button_menu.clear()

    def _set_enum_class(self, enum: enum.EnumMeta | None):
        """Set enum class from which members value should be selected."""
        if enum == self._enum_class:
            return None
        self._enum_class = enum
        self.clear()
        for i in self._enum_class.__members__.values():
            action = gui.Action()
            action.set_text(i.name.replace("_", " "))
            action.setData(i)
            self._action_map[i] = action
            action.setCheckable(True)
            self.button_menu.add(action)

    def get_enum_class(self) -> enum.EnumMeta | None:
        """Return current Enum class."""
        return self._enum_class

    def get_value(self) -> enum.Flag:
        """Current value as Enum member."""
        if self._enum_class is None:
            return None
        flag = self._enum_class(0)
        for k, v in self._action_map.items():
            if v.isChecked():
                flag |= k
        return flag

    def set_value(self, value: enum.Flag) -> None:
        """Set value with Enum."""
        if not isinstance(value, enum.Flag):
            value = self._enum_class(value)
        self._set_enum_class(value.__class__)
        if not isinstance(value, self._enum_class):
            raise TypeError(
                "setValue(self, Enum): argument 1 has unexpected type "
                f"{type(value).__name__!r}"
            )
        for i in list(value):
            self._action_map[i].setChecked(True)
        self.set_text(" | ".join(i.name for i in value))

    value = core.Property(enum.Flag, get_value, set_value, user=True)
    # enumClass = core.Property(type(enum.Flag), get_enum_class, set_enum_class)


if __name__ == "__main__":
    app = widgets.app()
    flag = widgets.AbstractItemView.EditTrigger
    w = EnumFlagWidget()
    w.set_value(flag.DoubleClicked)
    w.value_changed.connect(print)
    w.show()
    app.main_loop()



from __future__ import annotations

from prettyqt import gui, eventfilters


class ClickableLabelEventFilter(eventfilters.BaseEventFilter):
    """Eventfilter which underlines text on hover."""

    def __init__(self, parent, **kwargs):
        super().__init__(parent, **kwargs)
        parent.setFont(gui.Font("Arial"))
        parent.setFixedHeight(24)
        parent.set_size_policy("minimum", "expanding")

    def eventFilter(self, source, event):
        match event.type():
            case event.Type.Enter:
                with source.edit_font() as font:
                    font.setUnderline(True)
                source.set_cursor("pointing_hand")
                source.update()
            case event.Type.Leave:
                with source.edit_font() as font:
                    font.setUnderline(False)
                source.unsetCursor()
        return False

    def setText(self, text: str):
        fm = gui.FontMetrics(self.parent().font())
        width = fm.width(text)
        self.parent().setFixedWidth(width + 18)
        super().setText(text)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    widget = widgets.Label("test", tool_tip="testus")
    ef = ClickableLabelEventFilter(widget)
    widget.installEventFilter(ef)
    widget.show()
    app.main_loop()
